# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import logging
from typing import Any

from ckms import core

from cbra.conf import settings


AUTOLOAD: bool = getattr(settings, 'USE_DEFAULT_KEYS', False)

DEFAULT_KEYS: list[tuple[str, str]] = [
    ('enc', 'APP_ENCRYPTION_KEY'),
    ('sig', 'APP_SIGNING_KEY'),
    ('pii', 'PII_ENCRYPTION_KEY'),
    ('idx', 'PII_INDEX_KEY')
]

keychain: core.Keychain = core.get_default_keychain()


def get_keyspec_from_env(
    name: str,
    varname: str
) -> dict[str, Any] | None:
    """Inspect the environment variables for the present of `varname`,
    assume that its content is a DSN, and parse it into the keychain.
    """
    if varname not in os.environ:
        return None
    params: dict[str, Any] = core.parse_dsn(os.environ[varname])
    version: str | None = os.environ.get(f'{varname}_VERSION')
    if version is not None:
        # TODO: This is basically only used by Google Cloud KMS
        # keys.
        params['version'] = version
    return {name: params}


async def init():
    """Inspects the environment variables to add standardized keys to the
    global keychain.
    """
    if not AUTOLOAD:
        return
    logger = logging.getLogger("uvicorn")
    for name, varname in DEFAULT_KEYS:
        spec = get_keyspec_from_env(name, varname)
        if spec is None:
            continue
        keychain.configure(spec)
        logger.info("Configured keychain from environment variable %s", varname)
    await keychain


"""Declares :class:`DependantRunner`."""
import typing

import fastapi
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from .dependencyreceiver import DependencyReceiver


class DependantRunner:
    """Exposes a interface to run functions and methods that inject dependencies
    using :class:`cbra.Dependency`.
    """
    __module__: str = 'cbra'

    async def run_dependant(self,
        request: fastapi.Request,
        handle: typing.Callable,
        body: typing.Union[bytes, dict, list, str] = b'',
        path: str = '/',
        *args, **kwargs
    ):
        """Run dependant `handle` and return the result."""
        dependant = get_dependant(path=path, call=handle)
        dependant.dependencies.extend(self._get_dependants(path))
        if isinstance(handle, DependencyReceiver):
            dependant.dependencies.append(
                self._get_dependant(handle.inject, path)
            )
        values, errors, tasks, response, _ = await solve_dependencies(
            request=request,
            dependant=dependant,
            body=body,
            dependency_overrides_provider=None
        )
        if errors or tasks or response.status_code:
            raise NotImplementedError(errors, tasks, response)
        kwargs.update(values)
        return await dependant.call(*args, **kwargs)

    def get_dependants(self) -> typing.List[typing.Callable]:
        """Return the list of additional dependants that must be solved
        when invoking the callable.
        """
        return []

    def _get_dependants(self, path: str):
        return [
            get_parameterless_sub_dependant(
                depends=fastapi.Depends(x),
                path=path
            )
            for x in self.get_dependants()
        ]

    def _get_dependant(self, func, path):
        return get_parameterless_sub_dependant(
            depends=fastapi.Depends(func),
            path=path
        )



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .clientassertiontype import ClientAssertionType


class IntrospectionRequest(pydantic.BaseModel):
    client_assertion_type: ClientAssertionType | None = pydantic.Field(
        default=None,
        title="Client assertion type",
        description=(
            "The authorization server supports the following client assertions "
            "to establish the identity of confidential OAuth 2.0 clients:\n\n"
            "`urn:ietf:params:oauth:client-assertion-type:jwt-bearer`\n\n"
            "Clients that have registered a public key sign a JWT using that "
            "key. The Client authenticates in accordance with **JSON Web Token "
            "(JWT) Profile for OAuth 2.0 Client Authentication and Authorization "
            "Grants** and **Assertion Framework for OAuth 2.0 Client Authentication "
            "and Authorization Grants**. The JWT must contain the following "
            "required Claim Values and may contain the following optional Claim "
            "Values:\n\n"
            "- `iss` - (Required) Issuer. This MUST contain the ``client_id`` of "
            "the OAuth Client.\n"
            "- `sub` - (Required) Issuer. This MUST contain the ``client_id`` of "
            "the OAuth Client.\n"
            "- `aud` - (Required) Value that identifies the authorization server "
            "as an intended audience. The authorization server must verify that "
            "it is an intended audience for the token. The Audience must be the "
            "URL of the authorization server's **Token Endpoint**.\n"
            "- `jti` - (Required) A unique identifier for the token, which can "
            "be used to prevent reuse of the token. These tokens must only be "
            "used once.\n"
            "- `exp` - (Required) Expiration time on or after which the ID "
            "Token must not be accepted for processing.\n"
            "- `iat` - (Optional) Time at which the JWT was issued."
        )
    )

    client_assertion: str | None = pydantic.Field(
        default=None,
        title="Client assertion",
        description=(
            "The client assertion used to authenticate a confidential OAuth 2.0 "
            "client. See `client_assertion_type` for the supported protocols and "
            "formats."
        )
    )

    token: str = pydantic.Field(
        default=...,
        title="Token",
        description=(
            "The token to introspect. Must be issued by this authozation "
            "server."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic


class Event(pydantic.BaseModel):
    __module__: str = 'cbra.ext.moneybird.models'

    administration_id: str = pydantic.Field(
        default=...,
        title="Administration ID",
        description=(
            "The Moneybird administration in which the event occurred."
        )
    )

    webhook_id: str = pydantic.Field(
        default=...,
        title="Webhook ID",
        description=(
            "Identifies the webhook that sent the message."
        )
    )

    webhook_token: str = pydantic.Field(
        default=...,
        title="Token",
        description=(
            "A pre-shared secret that is used to verify that the event "
            "was sent by Moneybird."
        )
    )

    state: str | None = pydantic.Field(
        default=None,
        title="State",
        description=(
            "Describes the current state of the object on which the "
            "event occurred."
        )
    )

    entity_id: str = pydantic.Field(
        default=...,
        title="Entity ID",
        description=(
            "An identifier of the entity on which the event "
            "occurred."
        )
    )

    entity_type: str = pydantic.Field(
        default=...,
        title="Entity type",
        description=(
            "The type of the entity where the event occurred."
        )
    )

    entity: dict[str, Any] = pydantic.Field(
        default=...,
        title="Entity",
        description=(
            "Contains the entity on which the event occurred. Determine it's "
            "type by inspecting the `entity_type` attribute."
        )
    )

    @classmethod
    def new(cls, *, name: str, event: str) -> type['Event']:
        """Create a new model for the given event `event`."""
        return type(name, (cls,), {
            'action': pydantic.Field(
                default=...,
                title="Action",
                description=f'Is always `{event}`.'
            ),
            'Config': type('Config', (cls.Config,), {
                'title': f'Moneybird{name}'
            }),
            '__annotations__': {
                'action': Literal[event] # type: ignore
            }
        })


"""Declares :class:`TokenExchangeHandler`."""
import datetime
import typing

import fastapi
import pydantic

from cbra.ext import ioc
from .exceptions import Error
from .exceptions import UnsupportedSubjectTokenType
from .exceptions import UnsupportedTokenTypeRequested
from .iopenauthorizationserver import IOpenAuthorizationServer
from .isubject import ISubject
from .requestcontext import RequestContext
from .requesthandler import RequestHandler
from .tokenexchangerequest import TokenExchangeRequest
from .tokenexchangeresponse import TokenExchangeResponse
from .tokentype import TokenType


class TokenExchangeHandler(RequestHandler):
    """The base class to implement a handler for an OAuth 2.0
    token exchange as specified in :rfc:`8693`.
    """
    __module__: str = 'cbra.ext.oauth2'
    response_model: typing.Type[pydantic.BaseModel] = TokenExchangeResponse

    @classmethod
    def as_handler(cls, server: IOpenAuthorizationServer, **kwargs):
        """Return a request handler that invokes the handler."""
        async def f(
            dto: TokenExchangeRequest = TokenExchangeRequest.as_dependant(),
            ctx: RequestContext = ioc.instance(RequestContext),
            handler: TokenExchangeHandler = ioc.instance(cls),
        ) -> TokenExchangeResponse:
            handler.setup(**kwargs)
            handler.now = datetime.datetime.utcnow()
            return await handler.handle(ctx, dto)
        f.__doc__ = cls.__doc__
        return f

    async def apply_audience_policy(
        self,
        ctx: RequestContext,
        request: TokenExchangeRequest,
        sub: ISubject,
        scope: typing.Set[str]
    ):
        """Applies the policy for the ``resource`` specified in the token
        exchange request.

        The `ctx` parameter is a :class:`~cbra.ext.oauth2.RequestContext`
        instance describing the metadata of the :class:`TokenRequest`
        `request`.

        The subject identified :attr:`TokenExchangeRequest.subject_token`
        is provided through the `sub` parameter.

        The `scope` parameter is the scope as requested by the client,
        represented as a :class:`set`. To narrow down the scope, call
        the :meth:`set.remove()` method.

        The :meth:`apply_resource_policy` method does not have a return
        value. It is expected to update the `ctx` or `scope` accordingly.
        """
        pass

    async def apply_resource_policy(
        self,
        ctx: RequestContext,
        request: TokenExchangeRequest,
        sub: ISubject,
        scope: typing.Set[str]
    ):
        """Applies the policy for the ``resource`` specified in the token
        exchange request.

        The `ctx` parameter is a :class:`~cbra.ext.oauth2.RequestContext`
        instance describing the metadata of the :class:`TokenRequest`
        `request`.

        The subject identified :attr:`TokenExchangeRequest.subject_token`
        is provided through the `sub` parameter.

        The `scope` parameter is the scope as requested by the client,
        represented as a :class:`set`. To narrow down the scope, call
        the :meth:`set.remove()` method.

        The :meth:`apply_resource_policy` method does not have a return
        value. It is expected to update the `ctx` or `scope` accordingly.
        """
        pass

    async def apply_subject_policy(
        self,
        ctx: RequestContext,
        request: TokenExchangeRequest,
        sub: ISubject,
        scope: typing.Set[str]
    ):
        """Applies the policy that is specific for the subject requesting
        the token exchange.

        The `ctx` parameter is a :class:`~cbra.ext.oauth2.RequestContext`
        instance describing the metadata of the :class:`TokenRequest`
        `request`.

        The subject identified :attr:`TokenExchangeRequest.subject_token`
        is provided through the `sub` parameter.

        The `scope` parameter is the scope as requested by the client,
        represented as a :class:`set`. To narrow down the scope, call
        the :meth:`set.remove()` method.

        The :meth:`apply_subject_policy` method does not have a return
        value. It is expected to update the `ctx` or `scope` accordingly.
        """
        pass

    def can_issue(self, token_type: str) -> bool:
        """Return a boolean indicating if the server can issue the specified
        `token_type`.
        """
        return False

    def get_supported_token_types(self) -> typing.Set[str]:
        """Return the set of token types that the server is able to
        issue.
        """
        return set()

    def get_supported_subject_token_types(self) -> typing.Set[str]:
        """Return the set of subject token types supported."""
        return set()

    async def create_token_response(
        self,
        ctx: RequestContext,
        request: TokenExchangeRequest,
        sub: ISubject,
        scope: typing.Set[str]
    ) -> TokenExchangeResponse:
        """Invoked when all permissions and policies are validated. Create a
        token and return a corresponding :class:`TokenExchangeResponse`
        instance.
        """
        raise NotImplementedError

    async def handle(
        self,
        ctx: RequestContext = ioc.instance(RequestContext),
        dto: TokenExchangeRequest = TokenExchangeRequest.as_dependant()
    ) -> TokenExchangeResponse:
        """Handle an **OAuth 2.0 Token Exchange** request and return the result.
        
        Verify that the ``subject_token`` parameter of the exchange request is
        valid. Proceed to invoke methods enforcing the policies for the current
        subject, resource and audience. Finally, if all checks pass, generate
        the requested token and return it to the client.

        Args:
            ctx (:class:`~cbra.ext.oauth2.RequestContext`): describes the context
                of the current request and maintains state.
            dto (:class:`~cbra.ext.oauth2.TokenExchangeRequest`): the parameters
                of the **OAuth 2.0 Token Exchange** request.

        Raises:
            :exc:`~cbra.ext.oauth2.UnsupportedTokenTypeRequested`: the
                ``requested_token_type`` parameter specified a token type
                that is not supported by the server.
            :exc:`~cbra.ext.oauth2.UnsupportedSubjectTokenType`: the token
                used to identify the subject (``subject_token``) is of a
                type that is not supported by the server.

        Returns:
            :class:`~cbra.ext.oauth2.TokenExchangeResponse`: a datastructure
                holding the exchanged token and additional parameters describing
                the token.
        """
        if dto.actor_token:
            raise Error(
                error="invalid_request",
                error_description="The use of actor tokens is not implemented by this server."
            )
        if not self.can_issue(dto.requested_token_type):
            raise UnsupportedTokenTypeRequested(
                requested=dto.requested_token_type,
                supported=self.get_supported_token_types()
            )
        sub = await self._verify_subject_token(dto.subject_token_type, dto.subject_token)
        scope = dto.scope

        assert sub is not None # nosec
        if dto.resource:
            await self.apply_resource_policy(ctx, dto, sub, scope)
        if dto.audience:
            await self.apply_audience_policy(ctx, dto, sub, scope)
        await self.apply_subject_policy(ctx, dto, sub, scope)
        return await self.create_token_response(ctx, dto, sub, scope)

    async def verify_subject_token(self, token_type: str, token: str):
        """Verifies the ``subject_token`` parameter of the token exchange request. This
        method is expected to return a datastructure describing the subject of the
        token.
        """
        raise NotImplementedError

    async def _verify_subject_token(self, token_type, token):
        supported = self.get_supported_subject_token_types()
        if token_type not in supported:
            raise UnsupportedSubjectTokenType(
                given=token_type,
                supported=supported
            )
        return await self.verify_subject_token(token_type, token)


"""Declares :class:`FilterOption`."""
from typing import Any

import fastapi
import fastapi.params


class FilterOption:
    """Declares a query parameter that may be used to apply a filtering
    criterion to the collection endpoint.
    """
    __module__: str = 'cbra.resource'
    alias: str
    attname: str
    description: str
    annotation: type[Any]
    title: str | None
    default: type[Any] | None = None

    def __init__(
        self,
        *,
        alias: str,
        description: str,
        title: str | None = None,
        annotation: type[Any] = str,
        attname: str | None = None,
        default: Any | None = None
    ):
        self.alias = alias
        self.attname = attname or alias
        self.description = description
        self.annotation = annotation
        self.title = title
        self.default = default

    def add_to_namespace(self, attrs: dict[str, Any]) -> None:
        """Add the :class:`FilterOption` instance to the namespace
        (attributes) for a class, prior to constructing it during
        type initialization.
        """
        annotations = attrs.setdefault('__annotations__', {})
        annotations[self.attname] = self.annotation
        attrs[self.attname] = self.as_query()

    def as_query(self) -> fastapi.params.Query:
        """Return the :class:`FilterOption` as an instance of
        :class:`fastapi.params.Query`.
        """
        return fastapi.Query(
            alias=self.alias,
            default=self.default,
            title=self.title or str.title(self.alias),
            description=self.description,
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet
from unimatrix.exceptions import ImproperlyConfigured

from cbra import Endpoint
from cbra.cors import AnonymousReadCorsPolicy
from cbra.cors import CorsPolicyType
from cbra.params import CurrentServer
from .jsonwebsignature import JSONWebSignature


class RequestEncryptionKeyEndpoint(Endpoint):
    __module__: str = 'ckms.ext.rek'
    cors_policy: type[CorsPolicyType] = AnonymousReadCorsPolicy
    kid: str
    issuer: str = CurrentServer
    method: str = 'GET'
    mount_path: str = '.well-known/request-encryption-key'
    tags: list[str] = ['Server metadata']
    response_description: str = (
        "Request Encryption Key Set (REKS) wrapped in a JSON Web Signature (JWS)"
    )
    response_model: type[JSONWebSignature] = JSONWebSignature
    response_model_exclude: list[str] = []
    summary: str = "Response Encryption Key Set (REKS)"
    ttl: int = 300
    with_options: bool = True

    @property
    def codec(self) -> PayloadCodec:
        return PayloadCodec(signing_keys=[x for x in self.signers])

    @property
    def jwks(self) -> JSONWebKeySet:
        return self.keychain\
            .tagged('request-encryption-key')\
            .as_jwks()

    @property
    def keychain(self) -> Keychain:
        return self.request.app.keychain

    @property
    def signers(self) -> Keychain:
        return self.keychain\
            .tagged('reks-signer')\
            .private() # type: ignore

    async def handle(self) -> JSONWebSignature:
        """Returns a JSON Web Key Set (JWKS) that contains the public key,
        that a client can use to encrypt the body of an HTTP request,
        wrapped in a JSON Web Signature (JWS).
        """
        if not bool(self.signers):
            raise ImproperlyConfigured(
                "No signing keys for the Request Encryption Key Set (REKS) are "
                "configured. Tag keys with the 'reks-signer' tag to make "
                "them available for signing."
            )
        return JSONWebSignature.parse_raw(
            await self.codec.encode(
                payload={
                    'iss': self.issuer,
                    'iat': self.now,
                    'nbf': self.now,
                    'exp': self.now + self.ttl,
                    'jwks': self.jwks.dict()
                },
                content_type="jwt+reks"
            )
        )


"""Declares :class:`DependencySpec`."""
import typing

import pydantic


class DependencySpec(pydantic.BaseModel):
    """The base class for all dependency specifications."""
    __module__: str = 'cbra.ext.ioc.models'
    type: str
    name: str

    async def resolve(self) -> typing.Any:
        """Resolve the dependency specified by the input parameters."""
        raise NotImplementedError



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio.exceptions
import functools
import logging
import ssl
from typing import Any
from typing import Callable

import fastapi
import httpcore
import httpx
from httpx import AsyncClient
from httpx import Response
from ckms.types import AuthorizationServerNotDiscoverable
from ckms.types import AuthorizationServerMisbehaves
from ckms.types import JSONWebKeySet
from ckms.types import ServerMetadata

from cbra.conf import settings
from cbra.exceptions import UpstreamServiceNotAvailable
from cbra.ext.oauth2.types import IntrospectionResponse
from cbra.ext.oauth2.types import TokenResponse
from cbra.utils import retry
from .credential import Credential
from .personaldatahandler import PersonalDataHandler
from .resource import Resource
from .serviceidentity import ServiceIdentity


RESOURCE_SERVERS: dict[str, Any] = getattr(settings, 'RESOURCE_SERVERS', {})


UpstreamFailure: UpstreamServiceNotAvailable = UpstreamServiceNotAvailable(5)
Retryables: list[type[BaseException]] = [
    asyncio.exceptions.CancelledError,
    httpx.ConnectTimeout,
    httpx.ReadTimeout,
    httpcore.ConnectTimeout,
    httpcore.ReadTimeout,
    ssl.SSLWantReadError
]


class ServiceClient:
    __module__: str = 'cbra.ext.service'
    credential: Credential | None
    encrypter: PersonalDataHandler
    http: AsyncClient
    identity: ServiceIdentity
    jwks: JSONWebKeySet | None
    logger: logging.Logger
    metadata: ServerMetadata
    server: str
    timeout: float
    resources: dict[str, Resource]
    booted: bool = False

    @staticmethod
    def current(request: fastapi.Request) -> 'ServiceClient':
        return request.app.client

    @staticmethod
    def boot(func: Callable[..., Any]) -> Callable[..., Any]:
        """Ensure that the client is booted and configured."""
        @functools.wraps(func)
        async def f(self: Any, *args: Any, **kwargs: Any) -> Any:
            if not self.booted:
                await self.on_boot()
                self.booted = True
            return await func(self, *args, **kwargs)
        return f

    def __init__(
        self,
        server: str,
        identity: ServiceIdentity,
        encrypter: PersonalDataHandler,
        logger: logging.Logger | None = None,
        timeout: float = 60.0
    ):
        self.credential = None
        self.encrypter = encrypter
        self.identity = identity
        self.logger = logger or logging.getLogger("uvicorn")
        self.resources = {}
        self.server = server
        self.timeout = timeout

    @boot
    async def configure(self, resource: str) -> Resource:
        """Configure the named resource `resource` and return a
        :class:`~cbra.ext.service.Resource` instance.
        """
        if resource not in RESOURCE_SERVERS:
            raise ValueError(f"Resource not configured: {resource}")
        if resource not in self.resources:
            url = RESOURCE_SERVERS[resource]['server']
            self.resources[resource] = Resource(
                server=url,
                credential=await self.get_credential(
                    scope=RESOURCE_SERVERS[resource].get('scope') or set(),
                    resource=url
                )
            )
            await self.resources[resource].connect()
            self.logger.info("Configured resource server %s", url)
        return self.resources[resource]

    @boot
    async def get_client_assertion(self, audience: str | None = None) -> str:
        """Encode a client assertion used to authenticate with the authorization
        server.
        """
        return await self.identity.get_client_assertion(
            token_endpoint=audience or self.metadata.token_endpoint
        )

    @boot
    async def get_credential(
        self,
        scope: set[str],
        resource: str | None = None
    ) -> Credential:
        """Obtain a credential for the given resource."""
        return await Credential.obtain(
            http=self.http,
            token_endpoint=self.metadata.token_endpoint,
            client_id=self.identity.client_id,
            assertion_factory=self.get_client_assertion,
            scope=scope,
            resource=resource
        )

    async def decrypt(self, *args: Any, **kwargs: Any) -> bytes | dict[str, Any] | str:
        """Encrypt data from storage purposes."""
        return await self.encrypter.decrypt(*args, **kwargs)

    async def encrypt(self, *args: Any, **kwargs: Any) -> str:
        """Encrypt data for storage purposes."""
        return await self.encrypter.encrypt(*args, **kwargs)

    @boot
    async def token(self, *, grant_type: str, **params: Any) -> TokenResponse:
        """Request a grant from the token endpoint of the configured
        authorization server using client credentials.
        """
        params.update({
            'grant_type': grant_type,
            'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': await self.get_client_assertion(),
            'client_id': self.identity.client_id
        })
        response = await self.http.post( # type: ignore
            url=self.metadata.token_endpoint,
            json=params
        )
        if response.status_code >= 400:
            if str.startswith(response.headers.get('Content-Type', ''), 'application/json'):
                dto = response.json()
                self.logger.critical(
                    "Unable to obtain token (error: %s, description: %s)",
                    dto.get('error', '')[:256],
                    dto.get('error_description', '')[:256]
                )
                raise UpstreamServiceNotAvailable
            else:
                response.raise_for_status()
        return TokenResponse.parse_obj(response.json())

    @retry(15, interval=1.0, exception=UpstreamFailure, only=Retryables)
    async def get_server_jwks(self) -> JSONWebKeySet | None:
        self.logger.info("Obtaining JWKS for server %s", self.server)
        return await self.metadata.get_jwks(self.http)

    @retry(15, interval=1.0, exception=UpstreamFailure)
    async def get_server_metadata(self) -> ServerMetadata:
        self.logger.info("Obtaining metadata for server %s", self.server)
        return await ServerMetadata.discover(self.http, self.server)

    async def on_boot(self):
        try:
            self.logger.info("Booting service client")
            self.http = await AsyncClient(timeout=self.timeout).__aenter__()
            self.metadata = await self.get_server_metadata()
            self.jwks = await self.get_server_jwks()
        except AuthorizationServerNotDiscoverable:
            self.logger.fatal("Unable to discover authorization server %s", self.server)
            raise
        self.logger.info("Succesfully discovered authorization server %s", self.server)
        if not self.metadata.token_endpoint:
            self.logger.critical(
                "The authorization server did not advertise a token endpoint."
            )
            raise AuthorizationServerMisbehaves

    @boot
    async def introspect(self, access_token: str) -> IntrospectionResponse:
        """Introspect the given access token using the Token Introspection
        Endpoint of the configured authorization server.
        """
        if not self.metadata.introspection_endpoint:
            raise NotImplementedError(
                "The authorization server does not support introspection."
            )
        if not self.credential:
            self.credential = await self.get_credential(
                scope={"oauth2.introspect"},
                resource=self.server
            )
        response = await self.http.post( # type: ignore
            url=self.metadata.introspection_endpoint,
            json={
                'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
                'client_assertion': await self.get_client_assertion(
                    audience=self.metadata.introspection_endpoint
                ),
                'client_id': self.identity.client_id,
                'token': access_token
            }
        )
        response.raise_for_status()
        return IntrospectionResponse.parse_obj(response.json())

    async def on_teardown(self):
        if self.booted:
            self.logger.info("Teardown service client")
            for resource in self.resources.values():
                await resource.close()
            await self.http.__aexit__(None, None, None)

    @boot
    @retry(15, interval=1.0, exception=UpstreamFailure, only=Retryables)
    async def request(
        self,
        method: str,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ):
        """Make a HTTP request to the specified resource and path."""
        if resource not in self.resources:
            await self.configure(resource)
        svc = self.resources[resource]
        response = await svc.request(
            method=method,
            path=path,
            headers=headers,
            **kwargs
        )
        error_code: str | None = response.headers.get('X-Error-Code')
        if error_code == 'TOKEN_EXPIRED':
            assert svc.credential is not None # nosec
            await svc.credential.refresh()
            response = await svc.request(
                method=method,
                path=path,
                headers=headers,
                **kwargs
            )
        return response

    async def delete(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a DELETE request to the specified resource."""
        return await self.request(
            method="DELETE",
            resource=resource,
            path=path,
            headers=headers,
            **kwargs
        )

    async def get(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a POST request to the specified resource."""
        return await self.request(
            method="GET",
            resource=resource,
            path=path,
            headers=headers,
            **kwargs
        )

    async def post(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a POST request to the specified resource."""
        return await self.request(
            resource=resource,
            method="POST",
            path=path,
            headers=headers,
            **kwargs
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Callable
from typing import Coroutine

import httpx
from cbra.ext.oauth2.types import TokenResponse


class Credential:
    """Represents an access token for a specific resource server."""
    access_token: str
    assertion_factory: Callable[..., Coroutine[None, None, str]]
    client_id: str
    expires_in: int
    http: httpx.AsyncClient
    logger: logging.Logger = logging.getLogger("uvicorn")
    resource: str | None
    scope: set[str]
    token_endpoint: str
    CannotObtainCredential: type[Exception] = type('CannotObtainCredential', (Exception,), {})

    @staticmethod
    async def grant(
        http: httpx.AsyncClient,
        token_endpoint: str,
        client_id: str,
        assertion: str,
        scope: set[str],
        resource: str | None = None
    ) -> TokenResponse:
        """Invoke the token endpoint to obtain an access token."""
        params: dict[str, str] = {
            'grant_type': 'client_credentials',
            'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': assertion,
            'client_id': client_id,
            'scope': str.join(' ', sorted(scope)),
        }
        if resource is not None:
            params['resource'] = resource
        response = await http.post(url=token_endpoint, json=params) # type: ignore
        if response.status_code != 200:
            Credential.logger.critical(
                "Received non-200 response from %s: %s",
                token_endpoint, response.content[:256]
            )
            raise Credential.CannotObtainCredential(response.content[:256])
        return TokenResponse.parse_obj(response.json())

    @classmethod
    async def obtain(
        cls,
        http: httpx.AsyncClient,
        token_endpoint: str,
        client_id: str,
        assertion_factory: Callable[..., Coroutine[None, None, str]],
        scope: set[str],
        resource: str | None = None
    ) -> 'Credential':
        Credential.logger.info("Obtaining credential for service %s", resource)
        response = await Credential.grant(
            http=http,
            token_endpoint=token_endpoint,
            client_id=client_id,
            assertion=await assertion_factory(),
            scope=scope,
            resource=resource
        )
        return cls(
            client_id=client_id,
            access_token=response.access_token,
            expires_in=response.expires_in,
            http=http,
            assertion_factory=assertion_factory,
            scope=scope,
            token_endpoint=token_endpoint,
            resource=resource
        )

    def __init__(
        self,
        *,
        client_id: str,
        access_token: str,
        expires_in: int,
        http: httpx.AsyncClient,
        token_endpoint: str,
        assertion_factory: Callable[..., Coroutine[None, None, str]],
        scope: set[str],
        resource: str | None = None
    ):
        self.access_token = access_token
        self.assertion_factory = assertion_factory
        self.client_id = client_id
        self.expires_in = expires_in
        self.http = http
        self.resource = resource
        self.scope = scope
        self.token_endpoint = token_endpoint

    async def refresh(self):
        """Refreshes the credential."""
        self.logger.info("Refreshing credential for service %s", self.resource)
        response = await self.grant(
            http=self.http,
            token_endpoint=self.token_endpoint,
            client_id=self.client_id,
            assertion=await self.assertion_factory(),
            scope=self.scope,
            resource=self.resource
        )
        self.access_token = response.access_token
        self.expires_in = response.expires_in

    def __str__(self) -> str:
        return self.access_token

    


"""Declares :class:`TokenResponse`."""
import pydantic

from .iresponse import IResponse


class TokenResponse(pydantic.BaseModel, IResponse):
    __module__: str = 'cbra.ext.oauth2.types'

    access_token: str = pydantic.Field(
        title="Access token",
        description="The access token issued by the authorization server.",
        example=(
            "eyJhbGciOiAiRWREU0EiLCAia2lkIjogIjMwNTYzZWE4NmFiYWRlMWVjNWY1NmY0M"
            "GIwNzcyZjVmIiwgInR5cCI6ICJhdCtqd3QifQ.eyJhdWQiOiAiaHR0cHM6Ly9sb2N"
            "hbGhvc3Q6ODAwMCIsICJjbGllbnRfaWQiOiAicHVibGljIiwgImV4cCI6IDE2NDk1"
            "MjY0MzUsICJpYXQiOiAxNjQ5NTI2MTM1LCAiaXNzIjogImh0dHBzOi8vbG9jYWxob"
            "3N0OjgwMDAiLCAianRpIjogImpzeVFUa1dkemxmdy1jUGR6dnFuQV92QjMzbWNtbn"
            "FUIiwgIm5iZiI6IDE2NDk1MjYxMzUsICJzY29wZSI6ICJyZWFkIHdyaXRlIiwgInN"
            "1YiI6ICJib2JAZXhhbXBsZS51bmltYXRyaXhvbmUuaW8ifQ.EDansI-1VepV-l5dJ"
            "79IZolPcE3r2tewZrS91srOiGIhLA9bustHC_htZ4x4nzm0l4RWlaLxq7oA3svYAR"
            "ZblPSoI7GICEprqEblCYT8THn3__Blyeev4Tz9iTc2xvrxrCrompWVGGl2FZRyLkI"
            "-9xIA"
        )
    )

    token_type: str = pydantic.Field(
        title="Token type",
        description=(
            "The type of the token issued.\n\n"
            "The access token type provides the client with the information "
            "required to successfully utilize the access token to make a "
            "protected resource request (along with type-specific "
            "attributes).  The client must not use an access token if it "
            "does not understand the token type."
        ),
        example="Bearer"
    )
    expires_in: int = pydantic.Field(
        title="Expires in",
        description=(
            "The lifetime in seconds of the access token.  For example, the "
            "value `3600` denotes that the access token will expire in one "
            "hour from the time the response was generated."
        ),
        example=3600
    )
    state: str | None = pydantic.Field(
        title="State",
        description=(
            "Present if the `state` parameter was present in the client "
            "authorization request.  The exact value received from the client."
        ),
        example="jG-7nvPJ1-rpvUaTfstRfeA2WRExie1l"
    )

    refresh_token: str | None = pydantic.Field(
        title="Refresh token",
        description="Present if the client allows refresh tokens."
    )

    id_token: str | None = pydantic.Field(
        title="OpenID Connect ID Token",
        description=(
            "An OpenID Connect ID Token if the `response_type` parameter "
            "in the **Authorization Code Flow** was set to `code id_token`, "
            "otherwise `id_token` is absent."
        )
    )

    def is_error(self) -> bool:
        return False


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .authorizationrequestclaims import AuthorizationRequestClaims
from .basegrant import BaseGrant
from .iclient import IClient
from .isubject import ISubject
from .oidcrequestedclaims import OIDCRequestedClaims
from .policyfailure import PolicyFailure


class IOpenIdTokenBuilder:
    __module__: str = 'cbra.ext.oauth2.types'
    claims_model: type[AuthorizationRequestClaims]

    def requests(self, scope: set[str]) -> set[str]:
        """Return the set of claims requested by the given scope and
        claims.
        """
        raise NotImplementedError

    async def build(
        self,
        *,
        signing_key: str,
        client: IClient,
        subject: ISubject,
        grant: BaseGrant,
        scope: set[str],
        request: Any | None = None,
        access_token: str | None = None
    ) -> str | None:
        """Build an OpenID Connect ID Token."""
        raise NotImplementedError

    async def enforce_claims(
        self,
        request: Any,
        client: IClient,
        subject: ISubject,
        claims: OIDCRequestedClaims | None
    ) -> list[PolicyFailure]:
        """Enforce that the claims and the requested values match the
        values mandated by the client or the subject attributes.
        """
        raise NotImplementedError

    def parse_claims(self, claims: dict[str, Any]) -> AuthorizationRequestClaims:
        return self.claims_model.parse_obj(claims)

    def validate_claims(self, claims: dict[str, Any]) -> bool:
        """Return a boolean indicating if the ``claims`` parameter of an
        authorization request is valid.
        """
        try:
            self.claims_model.parse_obj(claims)
            return True
        except (pydantic.ValidationError, TypeError, ValueError):
            raise
            return False


"""Declares :class:`DependantSignature`."""
import collections
import inspect
import typing

import fastapi.params
import pydantic

from cbra.ext.ioc import Parameterless


REQUEST_DEPENDENCIES = (
    fastapi.params.Cookie,
    fastapi.params.Header,
    fastapi.params.Path,
    fastapi.params.Query,
)

REQUEST_UNION = typing.Union[REQUEST_DEPENDENCIES]


class DependantSignature:
    """Represents the signature of a dependant."""
    _parameters: typing.Dict[str, inspect.Parameter]
    _returns: typing.Any
    _signature: inspect.Signature

    @property
    def parameters(self) -> typing.Dict[str, inspect.Parameter]:
        """Return the list of signature parameters represented as
        :class:`inspect.Parameter` instances.
        """
        return self._parameters

    @property
    def returns(self) -> typing.Any:
        """The return type of the callable."""
        return self._returns

    def __init__(self, dependant: typing.Callable):
        self._signature = inspect.signature(dependant)
        self._parameters = collections.OrderedDict([
            (x.name, x) for x in list(self._signature.parameters.values())
        ])
        self._returns = self._signature.return_annotation

    def get_request_dependencies(
        self,
        as_parameters: bool = False
    ) -> typing.List[typing.Union[inspect.Parameter, Parameterless]]:
        """Return the list of dependencies that are resolved from a
        :class:`fastapi.Request` instance.
        """
        dependencies: typing.List[typing.Union[inspect.Parameter, Parameterless]] = []
        for param in self.parameters.values():
            if isinstance(param.default, REQUEST_DEPENDENCIES)\
            or inspect.isclass(param.annotation)\
            and issubclass(param.annotation, pydantic.BaseModel):
                p =Parameterless(
                    param.name,
                    param.default,
                    param.annotation
                )
                dependencies.append(p if not as_parameters else p.as_parameter())
            elif hasattr(param.default, 'get_request_dependencies'):
                dependencies.extend(param.default.get_request_dependencies())
            else:
                continue
        return dependencies



# pylint: skip-file
# type: ignore
import dataclasses
import datetime
import os
import typing

import ckms
import fastapi
import httpx
import pydantic

import cbra
from cbra.ext import oauth2
from cbra.ext.oauth2 import ref
from cbra.ext import jwks


@dataclasses.dataclass
class Subject(oauth2.types.ISubject):
    sub: str
    client_id: str


class ClientRepository(oauth2.types.IClientRepository):

    async def exists(self, client_id: str) -> bool:
        return True

    async def get(self, client_id: str) -> oauth2.ClientConfig:
        return oauth2.ClientConfig(
            client_id=client_id,
            default_redirect_url="https://localhost:8000",
            redirect_uris=["https://localhost:8000"],
            response_types={oauth2.types.ResponseType.code}
        )


class SubjectRepository(oauth2.types.ISubjectRepository):

    async def get(self, subject_id, client_id):
        return Subject(sub=subject_id, client_id=client_id)


class IdinName(pydantic.BaseModel):
    initials: str | None
    last_name: str | None
    last_name_prefix: str | None
    gender: str | None

    def get_name(self) -> str:
        return f'{self.initials} {self.get_family_name()}'

    def get_family_name(self) -> str:
        family_name = self.last_name
        if self.last_name_prefix is not None:
            family_name = f'{self.last_name_prefix} {family_name}'
        return typing.cast(str, family_name)

    def as_claims(self) -> dict[str, typing.Any]:
        claims = {
            'initials': self.initials,
            'name': self.get_name(),
            'family_name': self.get_family_name()
        }
        if self.gender in {'male', 'female'}:
            claims['gender'] = self.gender
        return claims


class IdinAge(pydantic.BaseModel):
    date_of_birth: datetime.date | None


class IdinTransaction(pydantic.BaseModel):
    transaction_id: str
    status: str
    issuer_id: str | None = None
    bin: str | None = None
    name: IdinName | None
    age: IdinAge | None

    def as_claims(self):
        return {
            **self.name.as_claims(),
            'birthdate': str(self.age.date_of_birth),
        }


class IdinProvider(oauth2.UpstreamProvider):
    name: str = 'idin'
    base_url: str = 'https://idin.cmtelecom.com/idin/v1.0'
    allowed_idin_issuers: set[str] = {"ASNBNL21", "BUNQNL2A"}

    async def create_redirect(
        self,
        request: fastapi.Request,
        dto: oauth2.types.AuthorizationRequest
    ) -> str:
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            if request.query_params.get('idin_issuer')\
            not in self.allowed_idin_issuers:
                raise oauth2.exceptions.Error(
                    error="invalid_request",
                    error_description="The given iDIN issuer was not valid.",
                    redirect_uri=dto.redirect_uri,
                    mode="redirect"
                )
            response = await client.post( # type: ignore
                url='/transaction',
                json={
                    'merchant_token': os.environ['CM_IDIN_MERCHANT_TOKEN'],
                    'entrance_code': dto.request_id,
                    'identity': True,
                    'name': True,
                    'gender': True,
                    'address': False,
                    'date_of_birth': True,
                    'issuer_id': request.query_params['idin_issuer'],
                    'merchant_return_url': self.reverse(request),
                    'language': 'nl'
                }
            )
            response.raise_for_status()
            result = response.json()
        dto.extra['idin_merchant_reference'] = result['merchant_reference']
        return result['issuer_authentication_url']

    def get_authorization_request_id(
        self,
        request: fastapi.Request
    ) -> str:
            return request.query_params['ec']

    async def process_response(
        self,
        request: fastapi.Request,
        params: oauth2.types.AuthorizationRequest
    ) -> None:
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post( # type: ignore
                url='/status',
                json={
                    'merchant_token': os.environ['CM_IDIN_MERCHANT_TOKEN'],
                    'merchant_reference': params.extra['idin_merchant_reference'],
                    'transaction_id': request.query_params['trxid'],
                }
            )
            result = IdinTransaction(**response.json())
            if result.status != 'success':
                raise oauth2.exceptions.Error(
                    error="consent_required",
                    error_description="The iDIN authentication failed.",
                    redirect_uri=params.redirect_uri,
                    mode='redirect'
                )
            print(result)


app = cbra.Application(keychain=ckms.Keychain())
app.add(
    ref.get_oauth2_server(
        grant_types={
            "authorization_code",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        },
        signing_key='sig'
    ),
    base_path='/oauth2'
)
app.keychain.register({
    'sig': {
        'provider': 'local',
        'path': 'pki/ed448.key',
        'tags': ['oauth2'],
        'use': 'sig'
    },
    'enc': {
        'provider': 'local',
        'path': 'pki/rsa-enc.key',
        'tags': ['oauth2'],
        'alg': "RSA-OAEP-256",
        'use': 'enc'
    },
})

if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



"""Declares :class:`ResponseType`."""
import enum


class ResponseType(str, enum.Enum):
    __module__: str = 'cbra.ext.oauth2.types'
    code = "code"
    code_id_token = "code id_token"
    code_id_token_token = "code id_token token"
    code_token = "code token"
    id_token = "id_token"
    id_token_token = "id_token token"
    none = "none"
    token = "token"


"""Declares :class:`JSONDefaultContentNegotiation`."""
from .default import DefaultContentNegotiation


class JSONDefaultContentNegotiation(DefaultContentNegotiation):
    """Content negotation that defaults to JSON. Used most commonly
    with misbehaving clients.
    """
    default_response_encoding: str  = "application/json"


"""Declares :class:`BaseCorsPolicy`."""
from typing import Any
from typing import NoReturn

import fastapi

from ..types import ICorsPolicy
from ..types import IRequestHandler


class BaseCorsPolicy(ICorsPolicy):
    """The base class for all Cross-Origin Resource Sharing (CORS)
    policy implementations.
    """
    __module__: str = 'cbra.cors'

    @classmethod
    def new(cls, **attrs: Any) -> type['BaseCorsPolicy']:
        return type(cls.__name__, (cls,), attrs)

    async def get_allowed_origins(self) -> set[str]:
        return self.allowed_origins

    async def process_request(
        self,
        request: fastapi.Request,
        handler: IRequestHandler,
        origin: str | None = None
    ) -> bool | NoReturn:
        """Determine if the origin is allowed to make cross-origin
        requests. Raise an exception if CORS is rejected for the
        given `origin`. If `origin` is ``None``, then it is assumed
        that the request is not cross-origin.
        """
        return True

    async def process_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response,
        handler: IRequestHandler,
        origin: str | None = None
    ) -> None | NoReturn:
        """Sets the appropriate headers on a HTTP response. If `origin`
        is not ``None``, then the origin was allowed to make cross-origin
        requests by this policy instance.
        """
        allow = await self.get_allowed_origins()
        if self.origin not in allow\
        or self.origin is None\
        or self.origin == self.server:
            return
        is_options = request.method == "OPTIONS"
        headers = {}

        # It is assumed here that this method is never executed if the
        # CORS request was rejected by process_request().
        headers['Access-Control-Allow-Origin'] = self.origin
        headers['Access-Control-Allow-Credentials'] = str(self.allow_credentials).lower()
        if self.allowed_response_headers and not is_options:
            headers['Access-Control-Expose-Headers'] = str.join(
                ', ',
                sorted(self.allowed_response_headers)
            )
        if self.allowed_methods:
            headers['Access-Control-Allow-Methods'] = str.join(
                ', ',
                sorted(self.allowed_methods)
            )
        if is_options:
            headers['Access-Control-Max-Age'] = str(self.max_age)
            if self.allowed_headers:
                headers['Access-Control-Allow-Headers'] = str.join(
                    ', ',
                    sorted(self.allowed_headers)
                )
        response.headers.update(headers) # type: ignore


"""Declares :class:`Configurable`."""
import importlib
import types
from typing import Any


class Configurable:
    __module__: str = 'cbra.ext.oauth2.types'
    kwargs_name: str = 'clients'
    settings_key: str = 'OAUTH_CLIENTS'

    @classmethod
    def configure(cls, **attrs: Any) -> type['Configurable']:
        return type(cls.__name__, (cls,), attrs)

    @classmethod
    def fromsettings(cls, settings: types.ModuleType) -> type['Configurable']:
        params = getattr(settings, cls.settings_key)
        if isinstance(params, str):
            params = {'class': params}
        module, symbol = str.rsplit(params.pop('class'), '.', 1)
        Configurable = getattr(importlib.import_module(module), symbol)
        return Configurable.configure(**params)


"""Declares :class:`SettingsSubjectRepository`."""
import types
from typing import Any

from cbra.conf import settings # type: ignore
from cbra.utils import docstring
from .types import BaseSubject
from .types import ISubject
from .types import ISubjectRepository


class SettingsSubjectRepository(ISubjectRepository):
    """A :class:`~cbra.ext.oauth2.types.ISubjectRepository` implementation
    that uses the :mod:`cbra` settings module as the source of the OAuth
    subjects.
    
    The settings module must declare the ``OAUTH_SUBJECTS`` attribute holding
    a dictionary that maps subject identifiers (i.e. its keys) to dictionaries
    specifying the additional properties of a subject.

    The subject model must be defined using the :attr:`model` attribute. The
    model defines which keys need to be declared for each subject.
    """
    __module__: str = 'cbra.ext.oauth2'
    settings: types.ModuleType = settings
    subjects: dict[int | str, Any]
    model: type[ISubject] = BaseSubject

    def __init__(self) -> None:
        self.subjects = {
            k: {**v, 'sub': k}
            for k, v in dict.items(getattr(self.settings, 'OAUTH_SUBJECTS', {})) # type: ignore
        }

    @docstring(ISubjectRepository)
    async def get(
        self,
        subject_id: int | str,
        client_id: str | None = None
    ) -> ISubject:
        return self.model.parse_obj({
            **self.subjects[subject_id],
            'client_id': client_id
        })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestProductFreeStockChanged(BaseMessageTest):

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': 'application/json',
                self.signature_header: "xmH72hlRe4aSLrI9eVAMOdbr8i7zdDoLkwg7Pl2E+3k="
            },
            'content': (
                b'{"idhook":12389,"name":"log-request-products-free_stock_changed","even'
                b't":"products.free_stock_changed","event_triggered_at":"2022-06-04 10:3'
                b'7:55","data":{"idproduct":19567274,"idvatgroup":13478,"idsupplier":nul'
                b'l,"productcode":"XXXTEST","name":"Test Product","price":0.08,"fixedsto'
                b'ckprice":0,"productcode_supplier":"","deliverytime":0,"description":"T'
                b'estproduct voor API- en systeemintegraties. Bestaat niet.","barcode":"'
                b'","unlimitedstock":false,"assembled":false,"type":"normal","weight":0,'
                b'"length":null,"width":null,"height":null,"minimum_purchase_quantity":0'
                b',"purchase_in_quantities_of":0,"hs_code":"","country_of_origin":"","ac'
                b'tive":true,"comment_count":0,"analysis_abc_classification":null,"analy'
                b'sis_pick_amount_per_day":null,"tags":{},"productfields":[{"idproductfi'
                b'eld":2603,"title":"Mapping 1: Categorie","value":"iPhone"},{"idproduct'
                b'field":2604,"title":"Mapping 2: Grading level","value":"ABC"},{"idprod'
                b'uctfield":2605,"title":"Mapping 3: Grade \/ Functionality","value":"A"'
                b'},{"idproductfield":2606,"title":"Mapping 5: Body Grade","value":"-"},'
                b'{"idproductfield":2607,"title":"Mapping 4: Screen Grade","value":"-"},'
                b'{"idproductfield":2608,"title":"Mapping 6: Device","value":"iPhone Izz'
                b'y"},{"idproductfield":2609,"title":"Mapping 7: Storage","value":"X"},{'
                b'"idproductfield":2610,"title":"Mapping 8: Release Year","value":"X"},{'
                b'"idproductfield":2611,"title":"Mapping 9: Macbook Inch","value":"X"},{'
                b'"idproductfield":2612,"title":"Mapping. 10: Parts Cat.","value":"X"},{'
                b'"idproductfield":2613,"title":"Mapping. 11: Accessoires Cat.","value":'
                b'"X"},{"idproductfield":2688,"title":"Mapping. 12: VAT","value":"VAT"},'
                b'{"idproductfield":2690,"title":"Mapping. 13: Colour","value":"-"}],"im'
                b'ages":[],"stock":[{"idwarehouse":6790,"stock":1,"reserved":1,"reserved'
                b'backorders":1,"reservedpicklists":0,"reservedallocations":0,"freestock'
                b'":0}],"pricelists":[]}}'
            )
        }


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any
from typing import Iterable

import pydantic
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import InvalidSignature
from ckms.utils import b64decode

from .iclient import IClient
from .irefreshtoken import IRefreshToken
from .isubject import ISubject


KEY_LENGTH: int = 64


class RefreshToken(pydantic.BaseModel, IRefreshToken):
    authorization_id: int
    id: int | None = None
    sub: int | str
    client_id: str
    secret: str
    created: datetime.datetime
    exchanged: datetime.datetime
    expires: datetime.datetime | None = None
    generation: int = 0
    scope: list[str]

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        values.setdefault('secret', secrets.token_urlsafe(KEY_LENGTH))
        return values

    @classmethod
    def new(
        cls,
        authorization_id: int,
        client_id: int | str,
        sub: int | str,
        scope: list[str] | set[str]
    ) -> 'RefreshToken':
        now = datetime.datetime.utcnow()
        return cls.parse_obj({
            'authorization_id': authorization_id,
            'client_id': client_id,
            'sub': sub,
            'secret': secrets.token_urlsafe(KEY_LENGTH),
            'created': now,
            'exchanged': now,
            'scope': scope
        })

    def allows_scope(self, scope: str | Iterable[str] | None) -> bool:
        """Return a boolean indicating if the Refresh Token allows the given
        scope.
        """
        scope = scope or self.scope
        if isinstance(scope, str):
            scope = {scope}
        return set(scope) <= set(self.scope)

    async def get_codec(self) -> PayloadCodec:
        keychain = Keychain()
        keychain.configure({
            'refresh_token': {
                'provider': 'local',
                'kty': 'oct',
                'alg': 'HS384',
                'use': 'sig',
                'key': {'cek': b64decode(self.secret)}   
            }
        })
        return PayloadCodec(
            verifier=await keychain,
            signing_keys=[keychain.get('refresh_token')]
        )

    async def generate(
        self,
        client: IClient,
        subject: ISubject,
        rotate: bool = False
    ) -> str | None:
        assert self.authorization_id is not None # nosec
        assert self.id is not None # nosec
        self.exchanged = datetime.datetime.utcnow()
        self.generation += 1
        if rotate:
            self.secret = secrets.token_urlsafe(KEY_LENGTH)
        return await client.create_refresh_token(
            codec=await self.get_codec(),
            subject=subject,
            authorization_id=self.authorization_id,
            token_id=self.id,
            expires=self.expires
        )

    async def verify(self, token: str) -> bool:
        """Return a boolean indicating if the given refresh token has a valid
        signature.
        """
        codec = await self.get_codec()
        try:
            await codec.decode(token, accept={"rt+jwt"})
            return True
        except InvalidSignature:
            return False


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from ckms.jose import PayloadCodec

from cbra.endpoint import Endpoint
from cbra.exceptions import Forbidden
from .currentclient import CurrentClient
from .params import ServerCodec
from .params import SubjectRepository
from .types import IClient
from .types import ISubjectRepository
from .types import IntrospectionRequest
from .types import IntrospectionResponse


class IntrospectionEndpoint(Endpoint):
    __module__: str = 'cbra.ext.oauth2'
    methods: list[str] = ["POST"]
    summary: str = "Introspection Endpoint"
    model: type[IntrospectionRequest] = IntrospectionRequest
    subjects: ISubjectRepository = SubjectRepository

    async def authorize(
        self,
        client: IClient = CurrentClient
    ) -> None:
        if not client.allows_scope({"oauth2.introspect"}):
            raise Forbidden

    async def handle( # type: ignore
        self,
        dto: IntrospectionRequest,
        codec: PayloadCodec = ServerCodec,
        client: IClient = CurrentClient,
    ) -> IntrospectionResponse:
        jws, jwt = await codec.jwt(dto.token, accept="at+jwt")
        subject = await self.subjects.get(
            client=client,
            subject_id=jwt.sub
        )
        if subject is None:
            return IntrospectionResponse(active=False)

        params: dict[str, Any] = {
            **jwt.dict(),
            'active': await jws.verify(codec.verifier)
        }
        if subject is not None:
            params.update({
                'sub': subject.sub
            })
        return IntrospectionResponse.parse_obj(params)


"""Declares :class:`Application`."""
import asyncio
import logging
import typing

import aorta
import fastapi
import unimatrix.runtime
from ckms.core import get_default_keychain
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from h11 import RemoteProtocolError
from starlette.middleware.base import BaseHTTPMiddleware
from unimatrix.exceptions import CanonicalException

from .apiroute import APIRoute
from .exceptions import BaseRedirectingException
from .exceptions import HTTPException
from .exceptions import TryAgain
from .exceptions import UnprocessableEntity
from .exceptions import UpstreamServiceNotAvailable
from .exceptions import UpstreamConnectionFailure
from .ext import ioc
from .types import IRequestHandler
from .types import IRouteable
from .types import Request
from .utils import retry


class Application(fastapi.FastAPI):
    logger: logging.Logger = logging.getLogger('uvicorn')

    #: The JWT codec used by the server for signing and encryption.
    codec: PayloadCodec

    #: The JSON Web Key Set (JWKS) offered by the server to encrypt
    #: data and verify signatures.
    jwks: JSONWebKeySet

    #: Tags to include in the JWKS.
    jwks_tags: list[str]

    #: The keychain holding the decryption and signing keys
    #: used by the server.
    keychain: Keychain

    #: A :class:`aorta.MessagePublisher` implementation that is used
    #: to publish messages.
    publisher: aorta.MessagePublisher

    def __init__(
        self,
        provider: ioc.Provider | None = None,
        log_exceptions: bool = False,
        jwks: JSONWebKeySet | None = None,
        jwks_tags: list[str] | None = None,
        keychain: Keychain | None = None,
        publisher: aorta.MessagePublisher | None = None,
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        """The :class:`cbra.Application` class integrates :mod:`cbra` with
        the underlying ASGI implementations (:mod:`fastapi` and :mod:`starlette`).

        The `provider` argument specifies the container to use when injecting
        dependencies. If `provider` is ``None`` (the default) then a new
        container is instantiated. In that case, the
        :attr:`cbra.conf.settings.DEPENDENCIES` is also evualated to inject
        dependencies based on the runtime settings. It is assumed that, if the
        dependency injection container is explicitely provided to the
        constructor, it does not need any further configuration.

        With `log_exceptions`, the application may be configured to explicitely
        log :exc:`cbra.exceptions.CanonicalException` instances. The default
        is ``False``.
        """
        self.jwks = jwks or JSONWebKeySet()
        self.jwks_tags = jwks_tags or ['server']
        self.keychain = keychain if keychain is not None else get_default_keychain()
        self.codec = PayloadCodec(
            decrypter=self.keychain,
            encrypter=self.keychain,
            signer=self.keychain,
            verifier=self.keychain
        )
        self.publisher = publisher or aorta.MessagePublisher(
            transport=aorta.transport.NullTransport()
        )
        on_startup = kwargs.setdefault('on_startup', [])
        on_startup.append(self.on_startup)
        on_shutdown = kwargs.setdefault('on_shutdown', [])
        on_shutdown.append(self._teardown)
        self.container = provider
        if self.container is None:
            self.container = ioc

        self.log_exceptions = log_exceptions
        kwargs.setdefault('docs_url', '/ui')
        kwargs.setdefault('openapi_tags', [])
        kwargs.setdefault('redoc_url', '/docs')
        exception_handlers = kwargs.setdefault('exception_handlers', {})
        exception_handlers.update({
            asyncio.TimeoutError: self.canonical_exception,
            BaseRedirectingException: self.canonical_exception,
            CanonicalException: self.canonical_exception,
            ConnectionError: self.canonical_exception,
            RequestValidationError: self.canonical_exception,
            TimeoutError: self.canonical_exception,
            RemoteProtocolError: self.canonical_exception,
            HTTPException: self.canonical_exception
        })
        super().__init__(*args, **kwargs)
        self.router.route_class = APIRoute
        self.add_middleware(BaseHTTPMiddleware, dispatch=self.handle_session)
        self.add_api_route(
            path='/.well-known/host-meta.json',
            endpoint=self.metadata
        )

    async def handle_session(
        self,
        request: Request,
        call_next: typing.Callable[..., typing.Coroutine[None, None, fastapi.Response]]
    ) -> fastapi.Response:
        response = await call_next(request)
        if request.scope.get('session'):
            await request.scope['session'].add_to_response(response)
        for handler in request.scope.get('handlers', []):
            await handler.on_response(request, response)
        return response

    def add(self,
        handler: typing.Union[
            IRouteable,
            type[IRequestHandler],
            type[IRouteable],
        ],
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        """Add a request handler to the :class:`Application`."""
        handler.add_to_router(app=self, *args, **kwargs)

    def add_api_route(self, *args: typing.Any, **kwargs: typing.Any): # type: ignore
        return super().add_api_route(*args, **kwargs)

    async def add_key(
        self,
        *,
        name: str,
        params: typing.Any
    ) -> None:
        """Add a key to the application and ensure that the key metadata
        is loaded.
        """
        self.keychain.configure({name: params})
        await self.keychain

    async def add_encryption_key(self, name: str, params: typing.Any) -> None:
        """Add a encryption key to the application and ensure that the key metadata
        is loaded.
        """
        params.update({
            'use': 'enc'
        })
        return await self.add_key(name=name, params=params)

    async def add_signing_key(self, name: str, params: typing.Any) -> None:
        """Add a signing key to the application and ensure that the key metadata
        is loaded.
        """
        params.update({
            'use': 'sig'
        })
        return await self.add_key(name=name, params=params)

    async def on_startup(self):
        await self._boot()

    async def _boot(self): # pragma: no cover
        self.logger.info("Booting up ASGI application")
        await unimatrix.runtime.on('boot')
        await self.setup_keychain()
        await ioc.on_boot(self)
        await self.boot()

    async def boot(self): # pragma: no cover
        """Invoked when the :class:`Application` is starting up. Subclasses
        may override this method to customize the behavior of the
        :class:`Application` during boot.

        Common use cases to override :meth:`boot()` is to establish database
        connections, fetch the public keys of consumed systems, or other
        run configurations that depend on active connections to other services.
        """
        pass

    async def canonical_exception(
        self,
        request: fastapi.Request,
        exception: BaseException
    ) -> typing.NoReturn | fastapi.Response:
        """Handles a canonical exception to a standard error message format."""
        origin = request.headers.get('Origin')
        if isinstance(exception, BaseRedirectingException):
            response = exception.as_response()
            if not response.headers.get('Access-Control-Allow-Origin') and origin:
                response.headers['Access-Control-Allow-Origin'] = origin
            return response
        if isinstance(exception, ConnectionRefusedError):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamServiceNotAvailable(**kwargs),
            )
        elif isinstance(
            exception,
            (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)
        ):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamConnectionFailure(**kwargs),
            )
        #elif isinstance(exception, UnsatisfiedDependency):
        #    return await self.canonical_exception(
        #        request, FeatureNotSupported()
        #    )
        elif isinstance(exception, RequestValidationError):
            raise
            return await self.canonical_exception(
                request=request,
                exception=UnprocessableEntity.fromexc(exception)
            )
        elif isinstance(exception, (asyncio.TimeoutError, TimeoutError)):
            #count = int(request.headers.get('X-Retry') or 0)
            return await self.canonical_exception(request, TryAgain(30))
        elif isinstance(exception, HTTPException):
            return await exception.as_response()
        elif isinstance(exception, CanonicalException):
            status_code = getattr(exception, 'http_status_code', None) or 500
            if self.log_exceptions:
                self.logger.error(
                    "%s (%s): %s",
                    exception.code,
                    exception.http_status_code,
                    exception.message
                )
                exception.log(self.logger.exception)
            response = JSONResponse(
                status_code=status_code,
                content=exception.as_dict(),
                headers={
                    **exception.get_http_headers(),
                    'X-Error-Code': exception.code
                }
            )

            # TODO: Ensure that a Javascript client can see the error response.
            # It is assumed here that if the client made a request that was able
            # to trigger an exception, it succesfully passed either a CORS
            # preflight check or it is a safe method, or if the CORS request
            # was not allowed it was rejected by the endpoint.
            if not response.headers.get('Access-Control-Allow-Origin') and origin:
                response.headers['Access-Control-Allow-Origin'] = origin
            return response
        else:
            raise NotImplementedError

    def get_jwks(self) -> JSONWebKeySet:
        return self.jwks

    def metadata(self, request: fastapi.Request, response: fastapi.Response):
        if request.headers.get('Origin'):
            response.headers['Access-Control-Allow-Origin'] = request.headers['Origin']
        return {
            'subject': str(request.url)
        }

    def openapi(self) -> dict[str, typing.Any]:
        if not self.openapi_schema:
            schema = super().openapi()

            # Remove some hardcoded stuff from endpoints where it is
            # not relevant, such as the 422 response for OPTIONS
            # endpoints.
            for path in schema['paths'].values():
                if 'options' in path and '422' in path['options']['responses']:
                    del path['options']['responses']['422']

            self.openapi_schema = schema
        return self.openapi_schema

    async def _teardown(self): # pragma: no cover
        self.logger.info("Tearing down ASGI application")
        await self.teardown()
        await ioc.on_shutdown(self)

    @retry(2, 0.01)
    async def setup_keychain(self) -> None:
        await self.keychain
        if self.jwks_tags:
            keychain = self.keychain.tagged(self.jwks_tags)
            self.jwks = keychain.as_jwks(private=False)

    async def teardown(self): # pragma: no cover
        """Invoked when the :class:`cbra.Application` is tearing down."""
        pass



"""Declares :class:`StringOrList`."""
import typing


class StringOrList(set):
    __module__: str = 'cbra.ext.oauth2'

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        data: typing.Union[str, typing.List[set]]
    ):
        if isinstance(data, str):
            data = [data]
        return set(sorted(data or []))

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: typing.Dict[str, typing.Any]
    ) -> None:
        field_schema.update(
            type="array",
            items={"type": "string"}
        )


"""Declares :class:`IApplication`."""
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet


class IApplication:
    codec: PayloadCodec

    def get_jwks(self) -> JSONWebKeySet:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
"""Declares :class:`WebhookEndpointMetaclass`."""
import copy
from collections import OrderedDict
from inspect import signature
from inspect import Parameter
from typing import cast
from typing import Any

from cbra import Endpoint
from cbra import EndpointMetaclass
from .models import WebhookResponse


class WebhookEndpointMetaclass(EndpointMetaclass):
    __module__: str = 'cbra.ext.webhook'

    def __new__(
        cls,
        name: str,
        bases: tuple[type[object]],
        attrs: dict[str, Any]
    ):
        Base = cast(
            type[Endpoint],
            super().__new__(cls, name, bases, attrs)
        )
        if Base.model is None or attrs.pop('__concrete__', False):
            return Base

        # Create a new class with the proper annotation
        class WebhookEndpoint(Base):
            __concrete__: bool = True

            async def handle(self, dto: Base.model) -> WebhookResponse:
                return await super().handle(dto)

        return WebhookEndpoint


"""Declares :class:`IAuthorizationServer`."""
from typing import Any

import fastapi

from .iupstreamprovider import IUpstreamProvider
from .servermetadata import ServerMetadata


class IAuthorizationServer:
    __module__: str = 'cbra.ext.oauth2.types'
    login_endpoint: str | None
    login_url: str | None
    metadata_endpoint: type[Any]
    providers: dict[str, IUpstreamProvider]

    def allows_grant(self, grant_type: str) -> bool:
        """Return a boolean indicating if the server is configured to allow
        the given grant type.
        """
        raise NotImplementedError

    async def get_metadata(self, request: fastapi.Request) -> ServerMetadata:
        """Return a datastructure describing the server endpoints
        and capabilities.
        """
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .error import Error
from .iresponse import IResponse
from .tokenresponse import TokenResponse


class TokenEndpointResponse(pydantic.BaseModel, IResponse):
    __root__: TokenResponse | Error

    @property
    def access_token(self) -> str:
        if not isinstance(self.__root__, TokenResponse):
            raise TypeError
        return self.__root__.access_token

    @property
    def error(self) -> str:
        if not isinstance(self.__root__, Error):
            raise TypeError
        return self.__root__.error

    @property
    def error_description(self) -> str | None:
        if not isinstance(self.__root__, Error):
            raise TypeError
        return self.__root__.error_description

    @property
    def refresh_token(self) -> str | None:
        if not isinstance(self.__root__, TokenResponse):
            raise TypeError
        return self.__root__.refresh_token

    def is_error(self) -> bool:
        return self.__root__.is_error()


"""Declares :class:`SymbolDependencySpec`."""
import typing
from typing import Any

import pydantic

from ..loader import import_symbol # type: ignore
from .dependencyspec import DependencySpec


class SymbolDependencySpec(DependencySpec):
    """Specifies the configuration format for a dependency that is loaded
    from the qualified name of a Python symbol.
    """
    __module__: str = 'cbra.ext.ioc.models'
    type: typing.Literal['symbol'] = 'symbol'
    qualname: str
    invoke: bool = False
    args: list[Any] = []
    kwargs: dict[str, Any] = {}

    async def resolve(self) -> typing.Any:
        """Resolve the dependency specified by the input parameters."""
        symbol = import_symbol(self.qualname) # type: ignore
        if self.invoke or self.args or self.kwargs:
            symbol = symbol(*self.args, **self.kwargs)
        return symbol



"""Declares :class:`Dependency`."""
import typing

import pydantic

from .filedependencyspec import FileDependencySpec
from .symboldependencyspec import SymbolDependencySpec


class Dependency(pydantic.BaseModel):
    """The specification of a dependency."""
    __module__: str = 'cbra.ext.ioc.models'
    force: typing.Optional[bool] = None
    spec: typing.Union[
        FileDependencySpec,
        SymbolDependencySpec
    ]

    async def resolve(self) -> typing.Any:
        """Resolve the dependency as defined in the specification."""
        return await self.spec.resolve()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

from .types import BaseGrant
from .types import AuthorizationRequest
from .types import IClient
from .types import ISubject
from .types import OIDCRequestedClaims


class OIDCClaimHandler:
    """Provides an interface to update a claims set for a given
    scope.
    """
    __module__: str = 'cbra.ext.oauth2'
    client: IClient
    context: Any
    handles: set[str] = set()
    initial: dict[str, Any]
    now: int
    requested: OIDCRequestedClaims
    subject: ISubject

    @classmethod
    def requests(cls, scope: set[str]) -> set[str]:
        return cls.get_requested_claims(scope)

    @classmethod
    def get_requested_claims(cls, scope: set[str]) -> set[str]:
        raise NotImplementedError(f'{cls.__name__}.get_requested_claims() must be overriden.')

    def __init__(
        self,
        context: Any,
        client: IClient,
        subject: ISubject,
        now: int,
        initial: dict[str, Any],
        claims: OIDCRequestedClaims | None = None
    ):
        self.requested = claims or OIDCRequestedClaims()
        self.client = client
        self.context = context
        self.initial = initial
        self.now = now
        self.subject = subject

    async def enforce(self, request: Any, claims: Any) -> None:
        """Enforce the claims requested as policy."""
        pass

    async def claims(
        self,
        grant: BaseGrant | None = None,
        request: AuthorizationRequest | None = None
    ) -> dict[str, Any] | None:
        """Return a mapping holding the claims that are added to the
        claims set by this handler.
        """
        raise NotImplementedError

    @functools.singledispatchmethod
    @classmethod
    def is_enabled(cls, config: set[str]) -> bool:
        """Return a boolean if the handler is enabled for the given
        scope or claims `config.
        """
        raise NotImplementedError(type(config))

    @is_enabled.register
    def _is_enabled_for_scope(self, scope: set) -> bool: # type: ignore
        return self.is_enabled_for_scope(scope) # type: ignore

    def is_enabled_for_scope(self, scope: set[str]) -> bool:
        """Return a boolean if the handler is enabled for the given
        scope `scope`.
        """
        return set(scope) >= set(self.handles)


"""Example webhook implementation for WooCommerce webhook
event messages.
"""
import cbra

from cbra.ext import woocommerce


class WooCommerceWebhook(woocommerce.WebhookEndpoint):
    echo: bool = True

    def get_hmac_secret(self) -> bytes:
        return b'foo'


def get_asgi_application():
    app = cbra.Application()
    app.add(WooCommerceWebhook)
    return app


app = get_asgi_application()
if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import pytest

from cbra import Application


@pytest.fixture
def app():
    return Application()


def random_key_name() -> str:
    return f'key-{secrets.token_hex(4)}'


@pytest.mark.asyncio
async def test_add_encryption_key(app: Application):
    name = random_key_name()
    await app.add_encryption_key(name, {
        'provider': 'local',
        'kty': 'OKP',
        'crv': 'Ed448'
    })
    meta = app.keychain.get(name)
    assert meta.use == 'enc'


@pytest.mark.asyncio
async def test_add_encryption_key_with_tags(app: Application):
    name = random_key_name()
    await app.add_encryption_key(name, {
        'provider': 'local',
        'kty': 'OKP',
        'crv': 'Ed448',
        'tags': ['foo']
    })
    meta = app.keychain.get(name)
    assert meta.use == 'enc'
    assert meta.tags == {'foo'}


@pytest.mark.asyncio
async def test_add_signing_key(app: Application):
    name = random_key_name()
    await app.add_signing_key(name, {
        'provider': 'local',
        'kty': 'OKP',
        'crv': 'Ed448'
    })
    meta = app.keychain.get(name)
    assert meta.use == 'sig'


@pytest.mark.asyncio
async def test_add_signing_key_with_tags(app: Application):
    name = random_key_name()
    await app.add_signing_key(name, {
        'provider': 'local',
        'kty': 'OKP',
        'crv': 'Ed448',
        'tags': ['foo']
    })
    meta = app.keychain.get(name)
    assert meta.use == 'sig'
    assert meta.tags == {'foo'}


"""Declares :class:`DefaultContentNegotiation`."""
from typing import Any

import fastapi

from cbra.renderers import NullRenderer
from cbra.types import IRenderer
from .default import DefaultContentNegotiation


class NullResponseContentNegotiation(DefaultContentNegotiation):
    """A content negotiation implementation that returns a response without a
    body.
    """
    __module__: str = 'cbra.negotiation'

    @classmethod
    def get_response_headers(cls) -> dict[str, Any]:
        return {}

    @classmethod
    def has_response_body(cls) -> bool:
        return False

    def select_renderer(
        self,
        renderers: list[type[IRenderer]],
        format_suffix: str | None = None,
        default: type[IRenderer] | None = None
    ) -> IRenderer:
        """Return a :class:`cbra.renderers.NullRenderer` instance."""
        return NullRenderer("*/*")


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from ckms.jose import PayloadCodec
from unimatrix.exceptions import CanonicalException

from .invalidrefreshtoken import InvalidRefreshToken


class RefreshTokenIdentifier(pydantic.BaseModel):
    authorization_id: int
    client_id: int | str
    id: int

    @classmethod
    def parse_jwt(
        cls,
        client_id: int | str,
        token: str
    ) -> 'RefreshTokenIdentifier':
        try:
            _, jwt = PayloadCodec.introspect(
                token=token,
                accept={"rt+jwt"}
            )
            if jwt is None:
                raise InvalidRefreshToken
            jwt.verify()
            authorization_id = int(jwt.extra.get('authorization_id')) # type: ignore
            token_id = int(jwt.iss) # type: ignore
        except (CanonicalException, ValueError, TypeError):
            raise InvalidRefreshToken
        return cls(
            client_id=client_id,
            id=token_id,
            authorization_id=authorization_id
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import functools
import logging
from typing import NoReturn
from typing import TypeAlias
from typing import Union

import fastapi
import fastapi.exceptions
import pydantic

from .exceptions import NotFound
from .exceptions import MissingRequestBody
from .exceptions import UncaughtException
from .exceptions import UnprocessableEntity



ExceptionType: TypeAlias = Union[
    fastapi.exceptions.RequestValidationError,
    pydantic.ValidationError
]

class ErrorHandler:
    """Handlers :mod:`pydantic` and :mod:`fastapi` errors."""
    __module__: str = 'cbra.errorhandler'
    logger: logging.Logger = logging.getLogger('uvicorn')
    request: fastapi.Request

    def __init__(
        self,
        request: fastapi.Request
    ):
        self.request = request

    @functools.singledispatchmethod
    async def handle(self, exception: ExceptionType) -> NoReturn:
        raise NotImplementedError(type(exception))

    @handle.register
    async def handle_request_validation_error(
        self,
        exception: fastapi.exceptions.RequestValidationError
    ) -> NoReturn:
        """Handles a :class:`fastapi.exceptions.RequestValidationError`
        instance.
        """
        field_errors: dict[tuple[int | str, ...], set[str]] = collections.defaultdict(set)
        for error in exception.errors():
            loc, *fieldpath = error.get('loc')
            fieldpath = tuple(fieldpath)
            typ = error.get('type')
            if loc in ('path', 'query'):
                raise NotFound
            elif loc == 'body' and fieldpath:
                field_errors[fieldpath].add(typ)
            elif loc == 'body' and typ == 'value_error.missing':
                raise MissingRequestBody

        if field_errors:
            raise UnprocessableEntity(errors=field_errors)
        self.logger.exception("Caught fatal %s", type(exception).__name__)
        raise UncaughtException


"""Declares :class:`DependencyReceiver`."""


class DependencyReceiver:
    """A mixin class that exposes a method to declare and receive dependencies
    injected through the :mod:`fastapi` dependency injection framework.
    """
    __module__: str = 'cbra'

    def inject(self) -> None:
        """Inject dependencies into the instance. To implement this feature,
        subclasses must override :meth:`inject()` and update the instance
        state accordingly. The default implementation does nothing.
        """
        pass



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .debugtransport import DebugTransport
from .aortaendpoint import AortaEndpoint
from .cloudschedulerendpoint import CloudSchedulerEndpoint
from .eventarcendpoint import EventarcEndpoint
from .googleendpoint import GoogleEndpoint
from .googletransport import GoogleTransport
from .service import Service
from .serviceaccountprincipal import ServiceAccountPrincipal


__all__: list[str] = [
    'AortaEndpoint',
    'CloudSchedulerEndpoint',
    'DebugTransport',
    'EventarcEndpoint',
    'GoogleEndpoint',
    'GoogleTransport',
    'Service',
    'ServiceAccountPrincipal'
]



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Iterable

from .refreshtoken import RefreshToken


class IAuthorization:
    __module__: str = 'cbra.ext.oauth2.types'

    def authorize(
        self,
        scope: str,
        now: datetime.datetime,
        expires: datetime.datetime | None = None
    ) -> None:
        """Authorizes the given scope `scope` for the Resource Owner
        and Client.
        """
        raise NotImplementedError

    def allows_refresh(self) -> bool:
        """Return a boolean indicating if the :class:`Authorization` allows
        the issue of a refresh token.
        """
        raise NotImplementedError

    def create_refresh_token(
        self,
        scope: Iterable[str]
    ) -> RefreshToken:
        """Create a *new* refresh token for the authorization."""
        raise NotImplementedError

    def enable_refresh(self) -> None:
        """Enables the issue of refresh tokens."""
        raise NotImplementedError

    def expire_scope(self, now: datetime.datetime | None = None) -> None:
        """Expire authorized scopes that have an expiration date."""
        raise NotImplementedError

    def get_active_scope(self, now: datetime.datetime | None = None) -> set[str]:
        """Return the set of scopes that are currently active."""
        raise NotImplementedError

    def is_authorized(self, scope: str | set[str]) -> bool:
        """Return a boolean indicating if authorization was granted previously
        to the given scope.
        """
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`DeletedEvent`."""
import pydantic


class DeleteEvent(pydantic.BaseModel):
    id: int = pydantic.Field(
        default=...,
        title="ID",
        description=(
            "The identifier of the object that was deleted. Use "
            "the `X-WC-Resource` header to determine the actual "
            "type of the resource."
        )
    )

    class Config:
        title: str = "WooCommerceDeleteEvent"


# type: ignore
"""Declares common exceptions classes."""
import copy
import typing
import fastapi
import urllib.parse
from typing import Any

import pydantic
import fastapi.responses
from fastapi.exceptions import RequestValidationError
from unimatrix.exceptions import CanonicalException
from unimatrix.exceptions import CanonicalExceptionMetaclass

from cbra.headers import CORS_HEADERS
from cbra.headers import DIGEST_SCHEMA
from cbra.types import ICanonicalException
from cbra.types import IContentNegotiation
from cbra.types import IParser
from cbra.types import IRenderer
from cbra.types import IRequestHandler
from cbra.utils import DeferredException



class HTTPException(Exception):
    """Mixin class for exceptions that can be converted to a response."""
    __module__: str = 'cbra.exceptions'

    async def as_response(self) -> fastapi.responses.Response:
        raise NotImplementedError


class IExceptionSpec(pydantic.BaseModel):
    code: str = pydantic.Field(
        default=...,
        title="Code",
        description=(
            "Further specifies the error condition in addition to the "
            "`kind` parameter."
        )
    )

    message: str = pydantic.Field(
        default=...,
        title="Message",
        description=(
            "A message describing the error. Note that this message "
            "is not intended for display to the end-user."
        )
    )

    hint: str | None = pydantic.Field(
        default=None,
        title="Hint",
        description=(
            "A hint describing possible actions to perform in order "
            "to resolve the error condition."
        )
    )


class BaseExceptionSpec(IExceptionSpec):
    detail: str

    @classmethod
    def for_code(
        cls,
        name: str,
        codes: list[str]
    ) -> typing.Any:
        return type(f'{name}Spec', (cls,), {
            'code': pydantic.Field(
                default=...,
                title="Error code",
                description=(
                    "A code describing the error that occurred. Clients "
                    "should use this code to determine the error condition."
                ),
                enum=codes
            ),
            '__annotations__': {
                'code': str
            },
            'Config': type('Config', (object,), {
                'title': f'{name}Spec'
            })
        })

    class Config:
        title = "ExceptionSpec"


class BaseRedirectingException(Exception):
    redirect_to: str
    params: dict[str, typing.Any] | None = None
    status_code: int = 303

    def __init__(
        self,
        redirect_to: str,
        params: dict[str, typing.Any] | None = None
    ) -> None:
        self.redirect_to = redirect_to
        self.params = params or {}

    def as_response(self) -> fastapi.responses.Response:
        return fastapi.responses.RedirectResponse(
            url=self.get_redirect_url(),
            status_code=self.status_code
        )

    def get_redirect_url(self) -> str:
        url = self.redirect_to
        if self.params:
            url = f'{url}?{urllib.parse.urlencode(self.params)}'
        return url

    def __repr__(self) -> str:
        return f"<LoginRequired: {self.get_redirect_url()}>"


class ValidationErrorSpec(BaseExceptionSpec):
    fieldErrors: typing.Dict[str, typing.List[typing.Any]] = {}

    class Config: # type: ignore
        title = "ValidationErrorSpec"


class ExceptionMetadata(pydantic.BaseModel):
    id: str = pydantic.Field(...,
        title="Identifier",
        description=(
            "Identifies the error. May be used to trace an error "
            "instance in logs."
        ),
        example="681b9604-df4e-4b2c-998c-9807a5569b1c"
    )

    timestamp: int = pydantic.Field(...,
        title="Timestamp",
        description=(
            "The date/time at which the error occurred, in milliseconds "
            "since the UNIX epoch."
        ),
        example=1647626707914
    )


class ExceptionModel(pydantic.BaseModel):
    apiVersion: typing.Literal['v1'] = pydantic.Field(...,
        title="API version",
        description="Schema version",
        example="v1"
    )
    metadata: ExceptionMetadata
    spec: BaseExceptionSpec

    @classmethod
    def parse_exc(
        cls,
        exception_class: type['CanonicalException'],
        data: dict[str, Any]
    ) -> 'CanonicalException':
        """Parse the exception data and return an instance of the given class."""
        obj = cls.parse_obj(data)
        return exception_class(id=obj.metadata.id, **obj.spec.dict())


class CanonicalExceptionMetaclass(CanonicalExceptionMetaclass):

    def __new__(cls: type, name: str, bases: list[str], attrs: dict[str, Any]):
        """Create a new :class:`CanonicalException` class."""
        new = super().__new__
        if attrs.pop('__abstract__', False):
            if name == 'CanonicalException':
                attrs['codes'] = {}
            return new(cls, name, bases, attrs)

        # Get the version and spec attributes to construct a pydantic model
        # description the exception.
        version: int = attrs.pop('version', 'v1')
        spec: type[BaseExceptionSpec] = attrs.pop('spec', BaseExceptionSpec)
        attrs['model'] = type(name, (ExceptionModel,), {
            '__annotations__': {'spec': spec},
            'spec': pydantic.Field(
                default=...,
                title=f'{name}Spec',
                description="Error condition parameters."
            )
        })
        return new(cls, name, bases, attrs)


class CanonicalException(CanonicalException, metaclass=CanonicalExceptionMetaclass):
    __abstract__: bool = True
    codes: dict[str, Any]
    model: type[ExceptionModel]
    include_in_schema: bool = True
    spec: type[IExceptionSpec] = BaseExceptionSpec

    @classmethod
    def defer(cls, **kwargs: typing.Any) -> DeferredException:
        return DeferredException(cls(**kwargs))

    @staticmethod
    def parse(code: str, data: dict[str, Any]) -> 'CanonicalException':
        for cls in CanonicalException.registered:
            if cls.code == code:
                break
        else:
            cls = None
        if cls is None or not hasattr(cls, 'spec'):
            return None
        return cls.model.parse_exc(cls, data)

    async def handle(
        self,
        request: fastapi.Request,
        handler: IRequestHandler,
        renderer: IRenderer
    ) -> fastapi.Response:
        """Renders the exception to a :class:`fastapi.responses.Response`
        instance.
        """
        if not renderer.has_content():
            return fastapi.Response(
                headers=self.get_http_headers(),
                status_code=self.http_status_code
            )

        return fastapi.Response(
            content=renderer.render(self.as_dict()),
            headers=self.get_http_headers(),
            status_code=self.http_status_code,
            media_type=renderer.get_response_encoding()
        )


class MissingScope(CanonicalException):
    __module__: str = 'cbra.exceptions'
    missing: set[str]
    http_status_code: int = 401
    code: str = "INVALID_SCOPE"
    message: str = "The access credential did not have the required scope."
    hint: str = (
        "Request the scope described by the 'missing' parameter from the "
        "authorization server."
    )

    class spec(IExceptionSpec):
        missing: list[str] = pydantic.Field(
            default=...,
            title="Required scope",
            description=(
                "The list of scopes that the principal needs to obtain "
                "to allow the request."
            )
        )

    def __init__(self, missing: set[str], **kwargs: Any):
        self.missing = missing
        super().__init__(**kwargs)

    def as_dict(self) -> dict[str, Any]:
        dto = super().as_dict()
        dto['spec']['missing'] = list(self.missing)
        return dto

    def upgrade_scope(self, scope: set[str]) -> set[str]:
        """Return a set holding the input scope and the required
        scope.
        """
        return set(self.missing) | scope


class UncaughtException(CanonicalException):
    http_status_code: int = 500


class Forbidden(CanonicalException):
    """Raised when the request is unconditionally rejected."""
    code = "FORBIDDEN"
    http_status_code = 403
    message = "The request was unconditionally rejected. That's all we know."


class WebhookMessageRejected(CanonicalException):
    """Raised when an incoming webhook message is rejected."""
    code: str = 'WEBHOOK_EVENT_REJECTED'
    http_status_code: int = 403
    message: str = "The server rejected the message."


class AuthenticationRequired(CanonicalException):
    """Raised when an operation required authenticated requests."""
    code = 'AUTHENTICATION_REQUIRED'
    http_status_code = 401
    message = "The operation requires authenticated requests."
    detail = (
        "An operation against a protected resource was attempted and it does "
        "not allow interactions with unauthenticated requests."
    )


class InvalidAuthorizationScheme(AuthenticationRequired):
    """Raised when the ``Authorization`` header had an invalid scheme."""
    code = 'INVALID_AUTHORIZATION_SCHEME'
    http_status_code = 403
    message = "The scheme provided in the Authorization header is not supported."
    detail = (
        "The server could not parse the credential in the Authorization header "
        "because the specified scheme is unknown or not allowed."
    )


class UntrustedIssuer(AuthenticationRequired):
    """Raised when a token was presented but the issuer is not trusted."""
    code: str = "UNTRUSTED_ISSUER"
    http_status_code: int = 403
    message: str = "The security token was issued by an untrusted party."

    def __init__(self, issuer: str):
        super().__init__(
            detail=(
                "The server has not established a trust relationship with "
                "the issuer ({issuer}) that supplied the security token. "
                "Consult the server documentation on the list of trusted "
                "issuers."
            )
        )


class MisbehavingIssuer(AuthenticationRequired):
    """Raised when an issuer is trusted, but there were problems retrieving
    its metadata or JSON Web Key Set (JWKS).
    """
    __module__: str = 'ckms.exceptions'
    code: str = 'MISBEHAVING_ISSUER'
    http_status_code: int = 403
    message: str = "The issuer is trusted but protocol errors occurred."


class UnreachableIssuer(AuthenticationRequired):
    """Raised when the issuer is trusted but no metadata could be discovered."""
    __module__: str = 'ckms.exceptions'
    code: str = 'UNREACHABLE_ISSUER'
    http_status_code: int = 403
    message: str = "The issuer is trusted but it was not reachable."


class ForgedAccessToken(AuthenticationRequired):
    """Raised when the parameters of a token were valid but it was not
    issued by a trusted entity.
    """
    code: str = "FORGED_ACCESS_TOKEN"
    http_status_code: int = 403
    message: str = "The security token was valid but not signed by a trusted party."
    detail: str = (
        "The claims presented with the security token validated against the server "
        "policy, but the digital signature did not validate against any of the known "
        "(public) keys. This incident is being logged, including your IP address "
        "and/or a browser fingerprint, and may lead to further escalation, "
        "such as criminal proscecution."
    )
    hint: str = "Stop forging security tokens."


class NotAuthorized(CanonicalException):
    """Raised when a request is not authorized to perform a given operation."""
    code = 'UNAUTHORIZED'
    http_status_code = 403
    message = "This request attempts an unauthorized action."
    detail = (
        "The credentials provided with the request did not resolve to a "
        "principal that was authorized to perform the action."
    )


class BearerAuthenticationRequired(AuthenticationRequired):
    hint = "Provide valid credentials using the Authorization header."


class NotFound(CanonicalException):
    code: str = 'NOT_FOUND'
    http_status_code: int = 404
    message: str = "No resource exists at the given URL."
    detail: str = (
        "The URL requested did not resolve to an existing resource."
    )


class InvalidPathParameter(NotFound):
    code: str = 'INVALID_PATH_PARAMETER'
    message: str = "The path parameter(s) did identify an existing resource."


class NotAcceptable(CanonicalException):
    """Generic exception that is raised when any of the ``Accept``,
    ``Accept-Encoding`` or ``Accept-Language`` headers can not be
    satisfied by the server.
    """
    __module__: str = 'cbra.exceptions'
    http_status_code: int = 406
    code: str = 'NOT_ACCEPTABLE'
    message: str = "The request's content negotiation is unacceptable."
    detail: str = (
        "The server cannot produce a response matching the list of "
        "acceptable values defined in the request's proactive content "
        "negotiation headers, and is unwilling to supply a default "
        "representation."
    )


class MediaTypeNotAcceptable(NotAcceptable):
    """Raised when the media type specified in the ``Accept`` header is
    not acceptable.
    """
    code: str = 'MEDIA_TYPE_NOT_ACCEPTABLE'


class UpstreamServiceNotAvailable(CanonicalException):
    """Raised when the application is not able to establish
    a (network) connection to an upstream service.
    """
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again later."
    retry_after: int | None = None

    def __init__(self, retry_after: int | None = None, **kwargs: Any):
        self.retry_after = retry_after
        if retry_after is not None:
            kwargs['hint'] = f'Try again in {retry_after} seconds.'
        super().__init__(**kwargs)

    def get_http_headers(self) -> dict[str, str]:
        return  {
            **super().get_http_headers(),
            'Retry-After': str(self.retry_after or 120)
        }


class UpstreamConnectionFailure(UpstreamServiceNotAvailable):
    """Raised when an upstream service listens at the configured address
    and port, but there are issued in establish the connection according
    to the agreed protocol. Such errors may occur when, for example, the
    upstream service is booting and has bound to its address and port, but
    is not yet ready to serve.
    """
    hint = "Try again in 10 seconds."


class TrustIssues(CanonicalException):
    http_status_code = 403
    code = "TRUST_ISSUES"
    message = (
        "The credential attached to the request was issued by an unkown "
        "authority."
    )
    hint = (
        "Verify that the credential (i.e. bearer token or X.509 certificate) "
        "was issued by an authority that is trusted by this system."
    )


class TryAgain(CanonicalException):
    code = 'SERVICE_NOT_AVAILABLE'
    http_status_code = 503
    message = "The service is currently not available."
    detail = (
        "Network or other infrastructure issues prevent "
        "proper operation of the service."
    )
    hint = "Try again in 600 seconds."

    def __init__(self, retry_after): # pragma: no cover
        super().__init__()


class UnsupportedMediaType(CanonicalException):
    http_status_code = 415
    code = "UNSUPPORTED_MEDIA_TYPE"
    message = "The server does not understand the request media type."

    @classmethod
    def fromparsers(cls, parsers: typing.List[typing.Type[IParser]]):
        return cls([x.media_type for x in parsers])

    def __init__(self, accepts: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.accepts = accepts

    def get_http_headers(self) -> dict:
        return {
            **super().get_http_headers(),
            'Accept': str.join(', ', self.accepts)
        }


class MissingContentTypeHeader(UnsupportedMediaType):
    __module__: str = 'cbra.exceptions'
    code: str = 'MISSING_CONTENT_TYPE_HEADER'
    message: str = (
        "The server does not understand the request media type because it "
        "did not provide the Content-Type header."
    )


class ParseError(CanonicalException):
    http_status_code = 400
    code = "MALFORMED_BODY"
    message = (
        "The server understands the media type but the request body "
        "is malformed."
    )


class InvalidHeaderValue(CanonicalException):
    http_status_code: int = 400
    code: str = "INVALID_HEADER_VALUE"

    def __init__(self, name: str, value: typing.Any = None, **kwargs: typing.Any):
        super().__init__(
            message=(
                f"The {name} header contained an illegal value."
                if value is not None else
                f"The {name} header contained an illegal value: '{value}'."
            )
        )


class UnsupportedDigestAlgorithm(CanonicalException):
    http_status_code: int = 400
    code: str = 'UNSUPPORTED_RESPONSE_DIGEST'
    message: str = (
        "None of the requested response digest algorithms is supported "
        "by the server."
    )
    hint: str = "Inspect the Wants-Digest header for the available algorithms."

    def __init__(self, supported: typing.Set[str]):
        super().__init__(
            detail=f"Supported algorithms are: {', '.join(supported)}."
        )
        self.supported = supported

    async def handle(
        self,
        request: fastapi.Request,
        handler: IRequestHandler,
        renderer: IRenderer
    ) -> fastapi.Response:
        response = await super().handle(request, handler, renderer)
        response.headers['Wants-Digest'] = ', '.join(self.supported)
        return response


class MissingRequestBody(CanonicalException):
    http_status_code: int = 422
    code: str = 'BODY_REQUIRED'
    message: str = "This endpoint requires a request body."
    detail: typing.Optional[str] = None
    hint: typing.Optional[str] = None


class UnprocessableEntity(CanonicalException):
    """Test"""
    http_status_code: int = 422
    code: str = "UNPROCESSABLE_ENTITY"
    message: str = "The server was not able to process the request."
    detail: str = (
        "The server understands the content type of the request entity, and "
        "the syntax of the request entity is correct, but it was unable to "
        "process the contained instructions."
    )
    hint: str = "Inspect the .spec.fieldErrors member for additional details."

    @classmethod
    def fromexc(
        cls,
        exception: typing.Union[pydantic.ValidationError, RequestValidationError]
    ):
        errors = {}
        for error in exception.errors():
            location = error['loc']
            if error['loc'][0] == 'body' and len(error['loc']) == 1:
                instance = MissingRequestBody()
                break
            elif location[0] == 'path':
                # A validation error in the path parameter suggests that
                # such a resource does not exist, assuming that the path
                # parameter is properly types.
                return InvalidPathParameter()
            elif exception.model:
                instance = cls()
            else:
                location, field = error['loc']
                if location not in errors:
                    errors[location] = []
        else:
            instance = cls(errors=errors)
        return instance

    def as_dict(self) -> dict[str, Any]:
        dto = super().as_dict()
        if self.params.get('errors'):
            dto['spec']['fieldErrors'] = {
                str.join('.', k): list(v) for k, v in self.params['errors'].items()
            }
        return dto


class Unserializable(ValueError):
    pass


def get_exception_headers(
    status_code: int
) -> typing.Dict[str, typing.Any]:
    return {
        **CORS_HEADERS,
        'Digest': DIGEST_SCHEMA,
        'X-Error-Code': {
            'description': "The error code describing the exception that occurred.",
            'schema': {
                'type': "string",
                'enum': sorted(set([
                    cls.code
                    for cls in CanonicalException.registered
                    if cls.http_status_code == status_code
                ]))
            }
        },
    }


def get_exception_models(
    status_code: int,
    name: str
) -> typing.TypeVar:
    codes = list(sorted(set([
        cls.code
        for cls in CanonicalException.registered
        if cls.http_status_code == status_code
    ])))
    return type(name, (ExceptionModel,), {
        'kind': pydantic.Field(
            default=...,
            title="Kind",
            description=f"Is always `{name}`."
        ),
        'spec': pydantic.Field(
            default=...,
            title="Specification",
            description="A datastructure describing the error."
        ),
        '__annotations__': {
            'kind': typing.Literal[name],
            'spec': BaseExceptionSpec.for_code(name, codes)
        }
    })


EXCEPTION_RESPONSES: dict[int | str, typing.Any] = {
    400: {
        'headers': get_exception_headers(400),
        'description': "The request is malformed.",
        'model': get_exception_models(400, 'BadRequest'),
    },
    401: {
        'headers': get_exception_headers(401),
        'description': (
            "Authentication or privilege escalation is required for this endpoint."
        ),
        'model': get_exception_models(401, 'NotAuthorized')
    },
    403: {
        'headers': get_exception_headers(403),
        'description': "The request was unconditionally rejected.",
        'model': get_exception_models(403, 'Forbidden')
    },
    406: {
        'headers': get_exception_headers(406),
        'description': (
            "The requested content representation is not supported by the server."
        ),
        'model': get_exception_models(406, 'NotAcceptable')
    },
    415: {
        'headers': get_exception_headers(415),
        'description': "The server does not understand the request media type.",
        'model': get_exception_models(415, 'UnsupportedMediaType')
    },
    422: {
        'headers': get_exception_headers(422),
        'description': "The content of the request was semantically invalid.",
        'model': get_exception_models(422, 'UnprocessableEntity')
    },
    500: {
        'headers': get_exception_headers(500),
        "description": (
            "An unrecoverable exception occurred while processing the request."
        ),
        'model': get_exception_models(500, 'InternalServerError')
    },
    503: {
        'headers': get_exception_headers(503),
        'description': (
            "The service or an upstream service is not available."
        ),
        'model': get_exception_models(503, 'ServiceNotAvailable')
    }
}


def get_exception_response(
    code: int | str,
    headers: dict[str, typing.Any],
    has_body: bool
) -> dict[str, typing.Any]:
    schema = copy.deepcopy(EXCEPTION_RESPONSES[code])
    if not has_body:
        schema.pop('model')
    return {
        **schema,
        'headers': headers
    }



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`CloudSchedulerEndpoint`."""
from typing import Any

from cbra.types import IContentNegotiation
from cbra.types import IRenderer
from cbra.utils import classproperty
from cbra.negotiation import NullContentNegotiation
from .googleendpoint import GoogleEndpoint


class CloudSchedulerEndpoint(GoogleEndpoint):
    """A :class:`~cbra.ext.google.GoogleEndpoint` implementation that handles
    a request that is received from Google Cloud Scheduler.
    """
    __module__: str = 'cbra.ext.google'
    negotiation: type[IContentNegotiation] = NullContentNegotiation
    default_response_code: int = 200
    renderers: list[type[IRenderer]] = []
    response_description: str = "The scheduled task completed succesfully."
    tags: list[str] = ["Cloud Scheduler"]

    @classproperty
    def responses(cls) -> dict[str | int, Any]:
        """The responses returned by this endpoint."""
        return {
            k: v for k, v in super().responses.items()
            if k in {400, 401, 403, 503, 'default'}
        }


# pylint: skip-file
import pytest

from ..filedependencyspec import FileDependencySpec


@pytest.mark.asyncio
async def test_resolve():
    spec = FileDependencySpec(
        name='ExampleDependency',
        path='VERSION'
    )
    with open(spec.path, 'rb') as f:
        assert await spec.resolve() == f.read()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi.responses

from cbra.exceptions import HTTPException


class PromptException(HTTPException):
    __module__: str = 'cbra.ext.oauth2.types'
    prompt_url: str
    parts: urllib.parse.ParseResult
    params: dict[str, str]

    def __init__(self, prompt_url: str, params: dict[str, str] | None = None) -> None:
        self.prompt_url = prompt_url
        self.parts = urllib.parse.urlparse(prompt_url)
        self.params = params or {}

    def get_redirect_uri(self) -> str:
        parts = list(self.parts)
        parts[4] = urllib.parse.urlencode(self.params, doseq=True)
        return urllib.parse.urlunparse(parts)

    async def as_response(self) -> fastapi.responses.Response:
        return fastapi.responses.RedirectResponse(
            url=self.get_redirect_uri(),
            status_code=303
        )


"""Declares :class:`BodyDependantModel`."""
import typing

import fastapi
import fastapi.params
import pydantic

from .dependantmodel import DependantModel


class BodyDependantModel(DependantModel):
    __module__: str = 'cbra.ext.oauth2'
    _source_class: typing.Type[pydantic.fields.FieldInfo] = fastapi.params.Body


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`OrderStatusType`."""
import enum


class OrderStatusType(str, enum.Enum):
    pending = 'pending'
    processing = 'processing'
    on_hold = 'on-hold'
    completed = 'completed'
    cancelled = 'cancelled'
    refunded = 'refunded'
    failed = 'failed'
    trash = 'trash'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class RefreshTokenPolicy(str, enum.Enum):
    """Describes the refresh policy for refresh tokens."""

    #: The token has a fixed expiration date, after which the resource
    #: owner must authorize the scope again.
    fixed = 'fixed'

    #: The token is valid for the number of seconds after it is issued,
    #: specified by the client.
    rolling = 'rolling'



# pylint: skip-file
from cbra.ext.oauth2 import ClientConfig
from cbra.ext.oauth2.types import GrantType
from cbra.ext.oauth2.types import ResponseMode


jwt: ClientConfig = ClientConfig(
    client_id='jwt',
    grant_types_supported=[
        GrantType.jwt_bearer.value
    ],
    response_modes=[
        ResponseMode.query
    ]
)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .oidcclaimrequest import OIDCClaimRequest


Empty = type('Empty', (object,), {})

#: OpenID Connect Core default ID Token claims.
DEFAULT_CLAIMS: set[str] = {
    "iss",
    "sub",
    "aud",
    "exp",
    "iat",
    "auth_time",
    "nonce",
    "acr",
    "amr",
    "azp"
}

class OIDCRequestedClaims(pydantic.BaseModel):
    iss: OIDCClaimRequest | None = None
    sub: OIDCClaimRequest | None = None
    aud: OIDCClaimRequest | None = None
    exp: OIDCClaimRequest | None = None
    iat: OIDCClaimRequest | None = None
    auth_time: OIDCClaimRequest | None = None
    nonce: OIDCClaimRequest | None = None
    acr: OIDCClaimRequest | None = None
    amr: OIDCClaimRequest | None = None
    azp: OIDCClaimRequest | None = None
    gender: OIDCClaimRequest | None = None

    def accepts(self, claim: str, value: Any) -> bool:
        """Return a boolean indicating if the given `value` is acceptable
        to the given `claim`.
        """
        if not self.requests(claim):
            return False
        return getattr(self, claim).accepts(value)

    def get_allowed_values(self, claim: str) -> list[Any]:
        """Return the list of allowed values for the given claim."""
        if not self.requests(claim):
            return []
        return getattr(self, claim).values

    def get_scope(self) -> set[str]:
        """Return a set of strings representing the requested scope by
        these claims.
        """
        scope: set[str] = {"openid"}
        for field in self.__fields__:
            assert isinstance(field, str) # nosec
            if self.requests(field):
                scope.add(field)
        return scope

    def requests(self, claim: str) -> bool:
        """Return a boolean indicating if the given claim is requested."""
        if claim not in self.__fields__:
            return False
        if claim in DEFAULT_CLAIMS:
            return True
        return getattr(self, claim, None) is not None


# pylint: skip-file
import base64
import hashlib
import typing

from ckms.utils import b64encode

from cbra.types import IResponseDigest


__all__ = [
    'SHA256Digest',
    'SHA512Digest',
]


class HaslibDigest(IResponseDigest):
    hashfunc: typing.Callable[[bytes], typing.Any]

    @classmethod
    def new(
        cls,
        name: str,
        func: typing.Callable[..., typing.Any] 
    ) -> typing.Type[IResponseDigest]:
        return type('HashlibDigest', (cls,), {
            'algorithm_name': name,
            'hashfunc': func
        })

    def calculate(self, message: bytes) -> str:
        digest = self.hashfunc(message).digest()
        return f'{self.algorithm_name}={bytes.decode(b64encode(digest), "ascii")}'


SHA256Digest = HaslibDigest.new('sha-256', hashlib.sha256)
SHA512Digest = HaslibDigest.new('sha-512', hashlib.sha512)


"""Declares :class:`Resource`."""
import typing
from inspect import Parameter
from typing import Any
from typing import Callable

from cbra.endpoint import Endpoint
from cbra.exceptions import get_exception_headers
from cbra.headers import CORS_HEADERS
from cbra.headers import DIGEST_SCHEMA
from cbra.headers import HEADERS_RESPONSE_BODY
from cbra.types import IEndpoint
from cbra.utils import classproperty
from .action import Action
from .filteroption import FilterOption
from .iresource import IResource
from .pathparameter import PathParameter
from .resourcemetaclass import ResourceMetaclass


class Resource(Endpoint, metaclass=ResourceMetaclass):
    __module__: str = 'cbra.resource'
    __abstract__: bool = True
    action: str
    default_limit: int = 100
    document: bool = True
    filter_actions: list[str] = ['list']
    filter_defaults: list[FilterOption] = [
        FilterOption(
            alias='offset',
            default=0,
            title="Offset",
            annotation=int,
            description="The offset applied to the search query."
        ),
        FilterOption(
            alias='limit',
            default=100,
            title="Limit",
            annotation=int,
            description=(
                "Limits the number of results included in the result set."
            )
        ),
        FilterOption(
            attname="order_by",
            alias='order-by',
            default=[],
            title="Order by",
            annotation=list[str],
            description="Specifies the ordering keys for the query."
        )
    ]
    filter_options: list[FilterOption] = []
    is_detail: bool = False
    parent: type[IResource] | None = None
    path_name: str | None = None
    path_parameter: typing.Union[str, PathParameter]
    require_authentication: bool = True
    subresources: list[type[Any]] = []
    verbose_name: str
    verbose_name_plural: str


    @staticmethod
    def expose(
        name: str,
        summary: str | None = None
    ) -> Callable[..., Any]:
        def decorator_factory(func: Callable[..., Any]):
            setattr(func, 'action', {
                'name': name,
                'summary': summary
            })
            return func
        return decorator_factory

    @classproperty
    def default_response(cls) -> typing.Dict[str, typing.Any]:
        return {
            'description': cls.response_description,
            'content': {
                x.media_type: {} for x in cls.renderers
            },
            'headers': {
                **CORS_HEADERS,
                **HEADERS_RESPONSE_BODY,
                'Digest': DIGEST_SCHEMA
            }
        }

    @classmethod
    def add_to_router( # type: ignore
        cls, *,
        app: IEndpoint.RouterType,
        base_path: str = '/',
        method: str = ...,
        request_handler: typing.Optional[
            typing.Callable[..., IEndpoint.ResponseTypes]
        ] = None,
        **kwargs: typing.Any
    ) -> None:
        if request_handler is ...:
            return super().add_to_router(
                app=app,
                base_path=base_path,
                cors_policy=cls.cors_policy,
                method=method,
                operation_id=str.lower(f'{cls.pluralname}.{cls.action}'),
                include_in_schema=cls.document,
                **kwargs
            )

        # Construct the base path based on the resource plural name.
        # All paths are relative to this path.
        assert cls.pluralname is not None # nosec
        base_path = f"{str.rstrip(base_path, '/')}/{str.lower(cls.path_name or cls.pluralname)}"

        cls = typing.cast(typing.Type[IResource], cls)
        for action in Action.fromclass(cls):
            action.add_to_router(app=app, base_path=base_path)

        # If there are subresources, the parent must define the
        # path_parameter attribute; subresources are always relative
        # to a detail URL.
        if cls.subresources:
            if not cls.path_parameter:
                raise ValueError(
                    f"For {cls.__module__}.{cls.__name__} to have "
                    f"subresources the {cls.__name__}.path_parameter "
                    "must be declared as a string value specifying "
                    "the parameter name."
                )

            child_mount: str = cls.path_parameter.get_path( # type: ignore
                base_path=str.rstrip(base_path, '/'),
                subpath=None,
                is_detail=True
            )

            # Construct subresource classes using the attributes from
            # the parent.
            for child in cls.subresources:
                if child.path_parameter == cls.path_parameter:
                    raise ValueError(
                        "Child path parameter must be different from parent."
                    )
                child.add_to_router(app=app, base_path=child_mount)

    @classmethod
    def get_path_signature(cls, detail: bool = False) -> list[Parameter]:
        params: list[Parameter] = []
        if detail or cls.is_detail:
            assert isinstance(cls.path_parameter, PathParameter)
            cls.path_parameter.insert_in(params)
        while cls.parent:
            cls = cls.parent
            assert isinstance(cls.path_parameter, PathParameter)
            cls.path_parameter.insert_in(params)
        return params

    def get_path_params(self) -> typing.Dict[str, typing.Any]:
        return typing.cast(
            typing.Dict[str, typing.Any],
            self.request.path_params # type: ignore
        )

    @classmethod
    def get_responses(cls) -> dict[int, typing.Any]:
        responses = super().get_responses()
        if cls.is_detail:
            responses[404] = {
                'headers': get_exception_headers(404),
                'description': (
                    "The resource specified by the path parameter(s) does not exist."
                )
            }
        return responses

    @classmethod
    def needs_parameter(cls, p: Parameter):
        needs_parameter = super().needs_parameter(p)
        return needs_parameter and (p.annotation != cls.model)

    @classmethod
    def update_signature(
        cls,
        runner: typing.Any,
        handle: typing.Any
    ) -> typing.Coroutine[typing.Any, typing.Any, IEndpoint.ResponseTypes]:
        runner = typing.cast(
            IResource,
            super().update_signature(runner, handle)
        )
        return runner

    def reverse(self, action: str, **params: Any) -> str:
        """Reverse the URL for the given action."""
        return self.request.url_for(str.lower(f'{self.name}.{action}'), **params)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
from cbra.ext import service

from .aortaendpoint import AortaEndpoint
from .commandendpoint import CommandEndpoint


class Service(service.Service):
    beat_prefix: str = '/.well-known/beat'

    def beat(
        self,
        name: str,
        command: type[aorta.Command],
        summary: str | None = None
    ) -> None:
        """Expose an endpoint where the scheduler can beat the application
        to produce the specified command.
        """
        self.add(CommandEndpoint.as_command_handler(name, command, summary))

    async def boot(self):
        await super().boot()
        self.add(AortaEndpoint)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi


HEADER_DELIVERY_ID: str = fastapi.Header(
    default=None,
    alias='X-WC-Webhook-Delivery-ID',
    title="Delivery ID",
    description=(
        "An identifier for this specific delivery. Use this value to lookup "
        "the delivery in the WooCommerce logs."
    )
)

HEADER_EVENT: str = fastapi.Header(
    default=None,
    alias='X-WC-Webhook-Event',
    title="Event",
    description=(
        "The event that occurred on the resource specified by the "
        "`X-WC-Webhook-Resource`, e.g. 'created', 'updated' or 'deleted'. "
    )
)

HEADER_SIGNATURE: str = fastapi.Header(
    default=None,
    alias='X-WC-Webhook-Signature',
    title="Signature",
    description=(
        "A Hash-based Message Authenticated Code (HMAC) calculated from "
        "the request body, using a previously shared key between "
        "the server and the WooCommerce instance. The signature is encoded "
        "using the Base64 encoding."
    )
)

HEADER_SOURCE: str  = fastapi.Header(
    default=None,
    alias="X-WC-Webhook-Source",
    title="Source",
    description=(
        "A URL specifing the source WooCommerce instance."
    )
)

HEADER_TOPIC: str = fastapi.Header(
    default=None,
    alias="X-WC-Webhook-Topic",
    title="Topic",
    description=(
        "The topic of the event, e.g. `order.updated`."
    )
)

HEADER_RESOURCE: str = fastapi.Header(
    default=None,
    alias="X-WC-Webhook-Resource",
    title="Resource",
    description=(
        "The kind of resource on which an event occurred."
    ),
    enum=['coupon', 'customer', 'order', 'product']
)

HEADER_WEBHOOK_ID: str = fastapi.Header(
    default=None,
    alias='X-WC-Webhook-ID',
    title="Webhook ID",
    description=(
        "Identifies the webhook that sent the event."
    )
)


"""Declares :class:`Parameterless`."""
import inspect
import typing

import fastapi.params

from .dependency import Dependency


class Parameterless(Dependency):
    """A :class:`Dependency` implementation that represents a parameterless
    depenency.
    """
    __module__: str = 'cbra.ext.ioc'

    @property
    def __signature__(self) -> inspect.Signature:
        return inspect.Signature([self.as_parameter()])

    def __init__(
        self,
        name: str,
        param: typing.Union[fastapi.params.Param, inspect._empty],
        annotation: typing.Optional[typing.Type[type]] = None
    ):
        self.annotation = annotation
        self.name = name
        self.param = param
        super().__init__(use_cache=True)

    def as_parameter(self) -> inspect.Parameter:
        return inspect.Parameter(
            name=self.name,
            kind=inspect.Parameter.KEYWORD_ONLY,
            default=self.param,
            annotation=self.annotation
        )

    async def resolve(self, **kwargs) -> typing.Any:
        return kwargs[self.name]

    def __repr__(self) -> str: # pragma: no cover
        return f"Parameterless({self.name}: {self.annotation.__name__})"



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`AortaEndpoint`."""
import asyncio
import warnings
from typing import Any
from typing import Awaitable

import aorta
import fastapi
from aorta.messagehandler import MessageHandler
from aorta.models import Message
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from cbra import Endpoint
from cbra.exceptions import Unserializable
from cbra.transaction import Transaction
from cbra.transaction import NullTransaction
from .eventarcendpoint import EventarcEndpoint
from .models import PubsubMessage


class AortaEndpoint(EventarcEndpoint):
    __module__: str = 'cbra.ext.google'
    summary: str = "Aorta"
    description: str = (
        "Receive a `google.cloud.pubsub.topic.v1.messagePublished` message "
        "containing an Aorta command, event or task, and run all handlers."
        "\n\nThis endpoint may respond with a non-2xx response code if any "
        "of the handlers fails to process the message."
    )
    provider: aorta.MessageHandlersProvider = aorta.get_default_provider()
    transaction: Transaction = fastapi.Depends(NullTransaction)
    mount_path: str = ".well-known/aorta"

    async def on_message(self, message: PubsubMessage) -> None | fastapi.Response:
        """Deserialize the Aorta object and run all matching handlers."""
        envelope = self.parse_aorta_message(message)
        handlers = self.provider.match(envelope) # type: ignore
        futures: list[Awaitable[None]] = []
        for handler_class in handlers:
            futures.append(self.run_handler(handler_class, envelope)) # type: ignore
        await asyncio.gather(*futures)

    async def run_handler(
        self,
        handler_class: type[MessageHandler],
        envelope: Message
    ) -> None:
        """Invoke `handler_class` with the given message."""
        obj: Any = envelope.get_object()
        dependant = get_parameterless_sub_dependant(
            depends=fastapi.Depends(handler_class),
            path='/'
        )
        values, errors, *_ = await solve_dependencies(
            request=self.request,
            dependant=dependant,
            body=obj,
            dependency_overrides_provider=None
        )
        if errors:
            warnings.warn(
                f"Improperly configured dependant: {handler_class.__name__}"
            )
            return
        handler: MessageHandler = dependant.call(**values) # type: ignore
        async with Transaction(publisher=self.publisher) as tx: # type: ignore
            tx.set_correlation_id(envelope.metadata.correlation_id)
            return await handler.handle(obj) # type: ignore

    def parse_aorta_message(self, message: PubsubMessage) -> Message:
        """Deserialize the message received from Google Pub/Sub and parse it
        into a previously registered Aorta type.
        """
        envelope: Message | None = None
        try:
            if not message.data:
                raise Unserializable
            envelope = self.provider.parse(message.get_data()) # type: ignore
        except Unserializable:
            warnings.warn(
                f"Message {message.message_id} did not contain an object "
                f"that was recognized by Aorta."
            )
            raise
        except aorta.UnknownMessageType:
            warnings.warn(
                f"Message {message.message_id} was succesfully deserialized "
                "but it did not contain an object that could be interpreted "
                "by Aorta."
            )
            raise

        self.logger.info(
            "Received %s/%s (id: %s, correlationId: %s)",
            envelope.api_version,
            envelope.kind,
            envelope.metadata.id,
            envelope.metadata.correlation_id
        )
        return envelope

    @Endpoint.register_exception # type: ignore
    async def on_unserializable(
        self,
        exception: Unserializable
    ) -> fastapi.Response:
        return fastapi.Response(
            status_code=200,
            headers={'X-Error-Code': "MALFORMED_OBJECT"}
        )

    @Endpoint.register_exception # type: ignore
    async def on_unknown_message_type(
        self,
        exception: aorta.exceptions.UnknownMessageType
    ) -> fastapi.Response:
        return fastapi.Response(
            status_code=200,
            headers={'X-Error-Code': "UNKNOWN_MESSAGE_TYPE"}
        )
